# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2016  Chao Liu <lcddzyx@gmail.com>

from __future__ import unicode_literals

import random
import logging

from . import fluxanalysis
from .expression import boolean

from six import iteritems, string_types

logger = logging.getLogger(__name__)


class ReactionDeletionStrategy(object):
    """Deleting reactions strategy class.

    When initializing instances of this class,
    :func:`.get_exchange_reactions` can be useful if exchange reactions
    are used as the test set.
    """

    def __init__(self, model, reaction_set=None):
        self._model = model
        if reaction_set is None:
            self._test_set = set(self._model.reactions)
        else:
            self._test_set = set(reaction_set)
        self._reactions = set(self._test_set)

    @property
    def entities(self):
        return self._reactions

    def iter_tests(self):
        while len(self._test_set) > 0:
            reaction = random.choice(tuple(self._test_set))
            self._test_set.remove(reaction)
            yield reaction, {reaction}

    def delete(self, entity, deleted_reactions):
        pass


class GeneDeletionStrategy(object):
    """Deleting genes strategy class.

    When initializing instances of this class, :func:`get_gene_associations`
    can be called to obtain the gene association dict from the model.
    """

    def __init__(self, model, gene_assoc):
        self._model = model
        self._genes = set()
        self._gene_assoc = dict(gene_assoc)

        for reaction, assoc in iteritems(self._gene_assoc):
            self._genes.update(v.symbol for v in assoc.variables)

        self._reactions = set(self._model.reactions)
        self._test_set = set(self._genes)

    @property
    def entities(self):
        return set(self._genes)

    def iter_tests(self):
        while len(self._test_set) > 0:
            gene = random.choice(tuple(self._test_set))
            self._test_set.remove(gene)

            deleted_reactions = set()
            self._new_gene_assoc = {}
            for reaction in self._reactions:
                if reaction not in self._gene_assoc:
                    continue
                assoc = self._gene_assoc[reaction]
                if boolean.Variable(gene) in assoc.variables:
                    new_assoc = assoc.substitute(
                        lambda v: v if v.symbol != gene else False)
                    if new_assoc.has_value() and not new_assoc.value:
                        deleted_reactions.add(reaction)
                    else:
                        self._new_gene_assoc[reaction] = new_assoc
                else:
                    self._new_gene_assoc[reaction] = assoc

            yield gene, deleted_reactions

    def delete(self, entity, deleted_reactions):
        self._reactions -= deleted_reactions
        self._gene_assoc = self._new_gene_assoc


def get_gene_associations(model):
    """Create gene association for class :class:`.GeneDeletionStrategy`.

    Return a dict mapping reaction IDs to
    :class:`psamm.expression.boolean.Expression` objects,
    representing relationships between reactions and related genes. This helper
    function should be called when creating :class:`.GeneDeletionStrategy`
    objects.

    Args:
        model: :class:`psamm.datasource.native.NativeModel`.
    """

    for reaction in model.reactions:
        assoc = None
        if reaction.genes is None:
            continue
        elif isinstance(reaction.genes, string_types):
            assoc = boolean.Expression(reaction.genes)
        else:
            variables = [boolean.Variable(g) for g in reaction.genes]
            assoc = boolean.Expression(boolean.And(*variables))
        yield reaction.id, assoc


def get_exchange_reactions(model):
    """Yield IDs of all exchange reactions from model.

    This helper function would be useful when creating
    :class:`.ReactionDeletionStrategy` objects.

    Args:
        model: :class:`psamm.metabolicmodel.MetabolicModel`.
    """
    for reaction_id in model.reactions:
        if model.is_exchange(reaction_id):
            yield reaction_id


def random_sparse(strategy, prob, obj_reaction, flux_threshold):
    """Find a random minimal network of model reactions.

    Given a reaction to optimize and a threshold, delete entities randomly
    until the flux of the reaction to optimize falls under the threshold.
    Keep deleting until no more entities can be deleted. It works
    with two strategies: deleting reactions or deleting genes (reactions
    related to certain genes).

    Args:
        strategy: :class:`.ReactionDeletionStrategy` or
            :class:`.GeneDeletionStrategy`.
        prob: :class:`psamm.fluxanalysis.FluxBalanceProblem`.
        obj_reaction: objective reactions to optimize.
        flux_threshold: threshold of max reaction flux.
    """

    essential = set()
    deleted = set()
    for entity, deleted_reactions in strategy.iter_tests():
        if obj_reaction in deleted_reactions:
            logger.info(
                'Marking entity {} as essential because the objective'
                ' reaction depends on this entity...'.format(entity))
            essential.add(entity)
            continue

        if len(deleted_reactions) == 0:
            logger.info(
                'No reactions were removed when entity {}'
                ' was deleted'.format(entity))
            deleted.add(entity)
            strategy.delete(entity, deleted_reactions)
            continue

        logger.info('Deleted reactions: {}'.format(
            ', '.join(deleted_reactions)))

        constr = []
        for r in deleted_reactions:
            flux_var = prob.get_flux_var(r)
            c, = prob.prob.add_linear_constraints(flux_var == 0)
            constr.append(c)

        logger.info('Trying FBA without reactions {}...'.format(
            ', '.join(deleted_reactions)))

        try:
            prob.maximize(obj_reaction)
        except fluxanalysis.FluxBalanceError:
            logger.info(
                'FBA is infeasible, marking {} as essential'.format(
                    entity))
            for c in constr:
                c.delete()
            essential.add(entity)
            continue

        logger.debug('Reaction {} has flux {}'.format(
            obj_reaction, prob.get_flux(obj_reaction)))

        if prob.get_flux(obj_reaction) < flux_threshold:
            for c in constr:
                c.delete()
            essential.add(entity)
            logger.info('Entity {} was essential'.format(
                entity))
        else:
            deleted.add(entity)
            strategy.delete(entity, deleted_reactions)
            logger.info('Entity {} was deleted'.format(entity))

    return essential, deleted


def random_sparse_return_all(strategy, prob, obj_reaction, flux_threshold):
    """Find a random minimal network of model reactions.

    Given a reaction to optimize and a threshold, delete entities randomly
    until the flux of the reaction to optimize falls under the threshold.
    Keep deleting until no more entities can be deleted. It works
    with two strategies: deleting reactions or deleting genes (reactions
    related to certain genes).

    Args:
        strategy: :class:`.ReactionDeletionStrategy` or
            :class:`.GeneDeletionStrategy`.
        prob: :class:`psamm.fluxanalysis.FluxBalanceProblem`.
        obj_reaction: objective reactions to optimize.
        flux_threshold: threshold of max reaction flux.
    """

    essential = set()
    deleted = set()
    essential_rxn = set()
    deleted_rxn = set()
    for entity, deleted_reactions in strategy.iter_tests():
        if obj_reaction in deleted_reactions:
            logger.info(
                'Marking entity {} as essential because the objective'
                ' reaction depends on this entity...'.format(entity))
            essential.add(entity)
            continue

        if len(deleted_reactions) == 0:
            logger.info(
                'No reactions were removed when entity {}'
                ' was deleted'.format(entity))
            deleted.add(entity)
            strategy.delete(entity, deleted_reactions)
            continue

        logger.info('Deleted reactions: {}'.format(
            ', '.join(deleted_reactions)))

        constr = []
        for r in deleted_reactions:
            flux_var = prob.get_flux_var(r)
            c, = prob.prob.add_linear_constraints(flux_var == 0)
            constr.append(c)

        logger.info('Trying FBA without reactions {}...'.format(
            ', '.join(deleted_reactions)))

        try:
            prob.maximize(obj_reaction)
        except fluxanalysis.FluxBalanceError:
            logger.info(
                'FBA is infeasible, marking {} as essential'.format(
                    entity))
            for c in constr:
                c.delete()
            essential.add(entity)
            for r in deleted_reactions:
                essential_rxn.add(r)
            continue

        logger.debug('Reaction {} has flux {}'.format(
            obj_reaction, prob.get_flux(obj_reaction)))

        if prob.get_flux(obj_reaction) < flux_threshold:
            for c in constr:
                c.delete()
            essential.add(entity)
            logger.info('Entity {} was essential'.format(
                entity))
            for r in deleted_reactions:
                essential_rxn.add(r)
        else:
            deleted.add(entity)
            strategy.delete(entity, deleted_reactions)
            logger.info('Entity {} was deleted'.format(entity))
            for r in deleted_reactions:
                deleted_rxn.add(r)

    return essential_rxn, deleted_rxn



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import logging

from six import itervalues

from ..command import Command
from psamm.datasource.native import reaction_signature

logger = logging.getLogger(__name__)


class DuplicatesCheck(Command):
    """Check for duplicated reactions in the model.

    This command reports any reactions in the model that appear more than
    once. Stoichiometry and reaction directionality is by default disregarded
    when checking for duplicates but can be enabled using the options.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--compare-direction', action='store_true',
            help='Take reaction directionality into consideration.')
        parser.add_argument(
            '--compare-stoichiometry', action='store_true',
            help='Take stoichiometry into consideration.')
        super(DuplicatesCheck, cls).init_parser(parser)

    def run(self):
        """Run check for duplicates"""

        # Create dictonary of signatures
        database_signatures = {}
        for entry in self._model.reactions:
            signature = reaction_signature(
                entry.equation, direction=self._args.compare_direction,
                stoichiometry=self._args.compare_stoichiometry)
            database_signatures.setdefault(signature, set()).add(
                (entry.id, entry.equation, entry.filemark))

        for reaction_set in itervalues(database_signatures):
            if len(reaction_set) > 1:
                print('Found {} duplicate reactions:'.format(
                    len(reaction_set)))
                for reaction, equation, filemark in reaction_set:
                    result = ' - {}: {}'.format(reaction, equation)
                    if filemark is not None:
                        result += ' (found in {})'.format(filemark)
                    print(result)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Importer for the COBRApy JSON format."""

import os
import glob
import json
import logging
import decimal

from six import iteritems

from ..reaction import Reaction, Compound, Direction
from ..datasource import native
from ..datasource.entry import (DictCompoundEntry as CompoundEntry,
                                DictReactionEntry as ReactionEntry,
                                DictCompartmentEntry as CompartmentEntry)

from ..importer import Importer as BaseImporter, ModelLoadError

logger = logging.getLogger(__name__)


def _float_parser(num_str):
    num = float(num_str)
    if num.is_integer():
        return int(num)
    else:
        return decimal.Decimal(num_str)


class Importer(BaseImporter):
    """Read metabolic model from COBRApy JSON format."""

    name = 'json'
    title = 'COBRApy JSON'
    generic = True

    def help(self):
        """Print help text for importer."""
        print('Source must contain the model definition in COBRApy JSON'
              ' format.\n'
              'Expected files in source directory:\n'
              '- *.json')

    def _resolve_source(self, source):
        """Resolve source to filepath if it is a directory."""
        if os.path.isdir(source):
            sources = glob.glob(os.path.join(source, '*.json'))
            if len(sources) == 0:
                raise ModelLoadError('No .json file found in source directory')
            elif len(sources) > 1:
                raise ModelLoadError(
                    'More than one .json file found in source directory')
            return sources[0]
        return source

    def _import(self, file):
        model_doc = json.load(file, parse_float=_float_parser)
        model = native.NativeModel()
        model.compartments.update(self._read_compartments(model_doc))
        model.compounds.update(self._read_compounds(model_doc))

        for reaction, lower, upper in self._read_reactions(model_doc):
            model.reactions.add_entry(reaction)
            if lower is not None and upper is not None:
                model.limits[reaction.id] = reaction.id, lower, upper

        model.name = model_doc.get('id', 'COBRA JSON model')

        biomass_reaction = None
        objective_reactions = set()
        for reaction in model_doc['reactions']:
            if reaction.get('objective_coefficient', 0) != 0:
                objective_reactions.add(reaction['id'])

        if len(objective_reactions) == 1:
            biomass_reaction = next(iter(objective_reactions))
            logger.info('Detected biomass reaction: {}'.format(
                biomass_reaction))
        elif len(objective_reactions) > 1:
            logger.warning(
                'Multiple reactions are used as the'
                ' biomass reaction: {}'.format(objective_reactions))

        model.biomass_reaction = biomass_reaction

        # Set compartment in reaction compounds
        compound_compartment = {}
        for compound in model.compounds:
            compartment = compound.properties.pop('compartment')
            compound_compartment[compound.id] = compartment

        for reaction in model.reactions:
            equation = reaction.equation
            if equation is None:
                continue

            # Translate compartments in reaction
            left = ((c.in_compartment(compound_compartment[c.name]), v) for
                    c, v in equation.left)
            right = ((c.in_compartment(compound_compartment[c.name]), v) for
                     c, v in equation.right)
            reaction.properties['equation'] = Reaction(
                equation.direction, left, right)

        return model

    def _read_compartments(self, doc):
        for compartment, name in iteritems(doc.get('compartments', {})):
            yield CompartmentEntry(dict(id=compartment, name=name))

    def _read_compounds(self, doc):
        for compound in doc['metabolites']:
            entry = CompoundEntry(compound)
            if 'formula' in entry.properties:
                formula = self._try_parse_formula(
                    entry.id, entry.properties['formula'])
                if formula is not None:
                    entry.properties['formula'] = formula
            yield entry

    def _parse_reaction_equation(self, entry):
        metabolites = entry.properties.pop('metabolites')
        compounds = ((Compound(metabolite), value)
                     for metabolite, value in iteritems(metabolites))
        if (entry.properties.get('lower_bound') == 0 and
                entry.properties.get('upper_bound') != 0):
            direction = Direction.Forward
        elif (entry.properties.get('lower_bound') != 0 and
              entry.properties.get('upper_bound') == 0):
            direction = Direction.Reverse
        else:
            direction = Direction.Both
        return Reaction(direction, compounds)

    def _read_reactions(self, doc):
        for reaction in doc['reactions']:
            entry = ReactionEntry(reaction)

            entry.properties['equation'] = (
                self._parse_reaction_equation(entry))

            lower, upper = None, None

            if 'lower_bound' in entry.properties:
                lower = entry.properties.pop('lower_bound')
            if 'upper_bound' in entry.properties:
                upper = entry.properties.pop('upper_bound')

            if 'gene_reaction_rule' in entry.properties:
                genes = self._try_parse_gene_association(
                    entry.id, entry.properties.pop('gene_reaction_rule'))
                if genes is not None:
                    entry.properties['genes'] = genes

            yield entry, lower, upper

    def import_model(self, source):
        """Import and return model instance."""
        if not hasattr(source, 'read'):  # Not a File-like object
            with open(self._resolve_source(source), 'r') as f:
                return self._import(f)
        else:
            return self._import(source)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2016  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

"""Module related to loading KEGG database files."""

import re
from types import FunctionType
from collections import Mapping

from .context import FileMark
from .entry import (CompoundEntry as BaseCompoundEntry,
                    ReactionEntry as BaseReactionEntry)
from ..reaction import Reaction, Compound, Direction
from ..expression.affine import Expression
from ..util import DictView

from six import iteritems, add_metaclass


class ParseError(Exception):
    """Exception used to signal errors while parsing"""


class KEGGEntry(object):
    """Base class for KEGG entry with raw values from KEGG."""

    def __init__(self, properties, filemark=None):
        self._properties = DictView(dict(properties))
        self._filemark = filemark

    @property
    def properties(self):
        return self._properties

    @property
    def filemark(self):
        return self._filemark


class _MappedEntry(object):
    """Base class for KEGG entry with values mapped to standard keys."""

    def __init__(self, mapper, entry):
        self._entry = entry
        self._properties = mapper(self._entry)

    @property
    def id(self):
        return self._properties['id']

    @property
    def properties(self):
        return self._properties

    @property
    def raw_entry(self):
        return self._entry

    @property
    def filemark(self):
        return self._entry.filemark


class CompoundEntry(_MappedEntry, BaseCompoundEntry):
    """KEGG entry with mapped compound properties."""
    def __init__(self, entry):
        super(CompoundEntry, self).__init__(CompoundMapper, entry)


class ReactionEntry(_MappedEntry, BaseReactionEntry):
    """KEGG entry with mapped reaction properties."""
    def __init__(self, entry):
        super(ReactionEntry, self).__init__(ReactionMapper, entry)


class _MapperMeta(Mapping.__class__):
    """Metaclass for mapper that turns methods into cached keys."""
    def __new__(cls, name, bases, namespace):
        property_accessors = {}
        new_namespace = {}
        for attr_name, attr in iteritems(namespace):
            if (isinstance(attr, FunctionType) and
                    not attr_name.startswith('_')):
                property_accessors[attr_name] = attr
            else:
                new_namespace[attr_name] = attr

        def getitem_props(self, key):
            if key in property_accessors:
                if key not in self._cache:
                    self._cache[key] = property_accessors[key](
                        self, self._entry.properties)
                return self._cache[key]
            raise KeyError('key not found: {!r}'.format(key))

        def iter_props(self):
            return iter(property_accessors)

        def len_props(self):
            return len(property_accessors)

        new_namespace['__getitem__'] = getitem_props
        new_namespace['__iter__'] = iter_props
        new_namespace['__len__'] = len_props

        return super(_MapperMeta, cls).__new__(
            cls, name, bases, new_namespace)

    def __call__(self, entry, *args, **kwargs):
        inst = super(_MapperMeta, self).__call__(*args, **kwargs)
        inst._entry = entry
        inst._cache = {}
        return inst


@add_metaclass(_MapperMeta)
class CompoundMapper(Mapping):
    """Mapper for raw KEGG compound properties to standard properties.

    Public methods are automatically translated into cached properties by the
    metaclass.
    """
    def id(self, raw):
        return raw['entry'][0].split(None, 1)[0]

    def name(self, raw):
        try:
            return next(self._iter_names(raw))
        except StopIteration:
            return None

    def names(self, raw):
        return list(self._iter_names(raw))

    def reactions(self, raw):
        return list(self._iter_reactions(raw))

    def enzymes(self, raw):
        return list(self._iter_enzymes(raw))

    def formula(self, raw):
        if 'formula' in raw:
            return raw['formula'][0]
        return None

    def exact_mass(self, raw):
        if 'exact_mass' in raw:
            return float(raw['exact_mass'][0])
        return None

    def mol_weight(self, raw):
        if 'mol_weight' in raw:
            return float(raw['mol_weight'][0])
        return None

    def pathways(self, raw):
        return list(self._iter_pathways(raw))

    def dblinks(self, raw):
        return list(self._iter_dblinks(raw))

    def comment(self, raw):
        if 'comment' in raw:
            return '\n'.join(raw['comment'])
        return None

    def _iter_names(self, raw):
        if 'name' in raw:
            for line in raw['name']:
                for name in line.rstrip(';').split(';'):
                    yield name.strip()

    def _iter_reactions(self, raw):
        if 'reaction' in raw:
            for line in raw['reaction']:
                for rxnid in line.split():
                    yield rxnid

    def _iter_enzymes(self, raw):
        if 'enzyme' in raw:
            for line in raw['enzyme']:
                for enzyme in line.split():
                    yield enzyme

    def _iter_pathways(self, raw):
        if 'pathway' in raw:
            for line in raw['pathway']:
                pathway, name = line.split(None, 1)
                yield pathway, name

    def _iter_dblinks(self, raw):
        if 'dblinks' in raw:
            for line in raw['dblinks']:
                database, entry = line.split(':', 1)
                yield database.strip(), entry.strip()


@add_metaclass(_MapperMeta)
class ReactionMapper(Mapping):
    """Mapper for raw KEGG reaction properties to standard properties.

    Methods are automatically translated into cached properties by the
    metaclass.
    """
    def id(self, raw):
        return raw['entry'][0].split(None, 1)[0]

    def name(self, raw):
        try:
            return next(self._iter_names(raw))
        except StopIteration:
            return None

    def names(self, raw):
        return list(self._iter_names(raw))

    def definition(self, raw):
        if 'definition' in raw:
            return raw['definition'][0]
        return None

    def equation(self, raw):
        if 'equation' in raw:
            return parse_reaction(raw['equation'][0])
        return None

    def enzymes(self, raw):
        return list(self._iter_enzymes(raw))

    def pathways(self, raw):
        return list(self._iter_pathways(raw))

    def comment(self, raw):
        if 'comment' in raw:
            return '\n'.join(raw['comment'])
        return None

    def rpairs(self, raw):
        return list(self._iter_rpairs(raw))

    def _iter_names(self, raw):
        if 'name' in raw:
            for line in raw['name']:
                for name in line.rstrip(';').split(';'):
                    yield name.strip()

    def _iter_enzymes(self, raw):
        if 'enzyme' in raw:
            for line in raw['enzyme']:
                for enzyme in line.split():
                    yield enzyme

    def _iter_pathways(self, raw):
        if 'pathway' in raw:
            for line in raw['pathway']:
                pathway, name = line.split(None, 1)
                yield pathway, name

    def _iter_rpairs(self, raw):
        if 'rpair' in raw:
            for line in raw['rpair']:
                pair, compounds, rp_type = line.split(None, 2)
                compounds = tuple(compounds.split('_', 1))
                yield pair, compounds, rp_type


def parse_kegg_entries(f, context=None):
    """Iterate over entries in KEGG file."""

    section_id = None
    entry_line = None
    properties = {}
    for lineno, line in enumerate(f):
        if line.strip() == '///':
            # End of entry
            mark = FileMark(context, entry_line, 0)
            yield KEGGEntry(properties, filemark=mark)
            properties = {}
            section_id = None
            entry_line = None
        else:
            if entry_line is None:
                entry_line = lineno

            # Look for beginning of section
            m = re.match(r'([A-Z_]+)\s+(.*)', line.rstrip())
            if m is not None:
                section_id = m.group(1).lower()
                properties[section_id] = [m.group(2)]
            elif section_id is not None:
                properties[section_id].append(line.strip())
            else:
                raise ParseError(
                    'Missing section identifier at line {}'.format(lineno))


def parse_compound_file(f, context=None):
    """Iterate over the compound entries in the given file."""
    for entry in parse_kegg_entries(f, context):
        yield CompoundEntry(entry)


def parse_reaction_file(f, context=None):
    """Iterate over the reaction entries in the given file."""
    for entry in parse_kegg_entries(f, context):
        yield ReactionEntry(entry)


def parse_reaction(s):
    """Parse a KEGG reaction string"""

    def parse_count(s):
        m = re.match(r'^\((.+)\)$', s)
        if m is not None:
            s = m.group(1)

        m = re.match(r'^\d+$', s)
        if m is not None:
            return int(m.group(0))

        return Expression(s)

    def parse_compound(s):
        m = re.match(r'(.+)\((.+)\)', s)
        if m is not None:
            return Compound(m.group(1), arguments=[Expression(m.group(2))])
        return Compound(s)

    def parse_compound_list(s):
        for cpd in s.split(' + '):
            if cpd == '':
                continue

            fields = cpd.strip().split(' ')
            if len(fields) > 2:
                raise ParseError(
                    'Malformed compound specification: {}'.format(cpd))
            if len(fields) == 1:
                count = 1
                compound = parse_compound(fields[0])
            else:
                count = parse_count(fields[0])
                compound = parse_compound(fields[1])

            yield compound, count

    cpd_left, cpd_right = s.split('<=>')
    left = parse_compound_list(cpd_left.strip())
    right = parse_compound_list(cpd_right.strip())

    return Reaction(Direction.Both, left, right)



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>
# Copyright 2020 Elysha Sameth <esameth1@my.uri.edu>


import unittest

from psamm import findprimarypairs
from psamm.formula import Formula, Atom, Radical
from psamm.reaction import Compound
from psamm.datasource.reaction import parse_reaction
from itertools import permutations


class TestFindPrimaryPairs(unittest.TestCase):
    def test_default_weight(self):
        self.assertEqual(findprimarypairs.element_weight(Atom.C), 1)
        self.assertEqual(findprimarypairs.element_weight(Atom.H), 0)
        self.assertEqual(findprimarypairs.element_weight(Atom.N), 0.82)
        self.assertEqual(findprimarypairs.element_weight(Radical('R')), 0.82)

    def test_predict_compound_pairs(self):
        """Test prediction of HEX1 reaction."""
        reaction = parse_reaction(
            'atp[c] + glc-D[c] => adp[c] + g6p[c] + h[c]')
        formulas = {
            'atp': Formula.parse('C10H12N5O13P3'),
            'adp': Formula.parse('C10H12N5O10P2'),
            'glc-D': Formula.parse('C6H12O6'),
            'g6p': Formula.parse('C6H11O9P'),
            'h': Formula.parse('H')
        }
        transfer, balance, ambiguous_pairs = \
            findprimarypairs.predict_compound_pairs(
                reaction, formulas)

        self.assertEqual(balance, {})
        self.assertEqual(transfer, {
            ((Compound('atp', 'c'), 1), (Compound('adp', 'c'), 1)):
                Formula.parse('C10H12N5O10P2'),
            ((Compound('glc-D', 'c'), 1), (Compound('g6p', 'c'), 1)):
                Formula.parse('C6H11O6'),
            ((Compound('atp', 'c'), 1), (Compound('g6p', 'c'), 1)):
                Formula.parse('O3P'),
            ((Compound('glc-D', 'c'), 1), (Compound('h', 'c'), 1)):
                Formula.parse('H')
        })
        self.assertEqual(ambiguous_pairs, set())

    def test_predict_compound_pairs_ambig(self):
        """Test prediction of (non-sense) ambiguous reaction."""
        reaction = parse_reaction(
            'a[c] + b[c] => c[c] + d[c]')
        formulas = {
            'a': Formula.parse('C10H10'),
            'b': Formula.parse('C10H10'),
            'c': Formula.parse('C10H10'),
            'd': Formula.parse('C10H10')
        }
        transfer, balance, ambiguous_pairs = \
            findprimarypairs.predict_compound_pairs(
                reaction, formulas)

        all_perm = set(set(permutations(list(ambiguous_pairs)[0])))
        ambig = list(ambiguous_pairs)[0]

        self.assertEqual(balance, {})
        self.assertEqual(transfer, {
            ((Compound('b', 'c'), 1), (Compound('d', 'c'), 1)):
                Formula.parse('C10H10'),
            ((Compound('a', 'c'), 1), (Compound('c', 'c'), 1)):
                Formula.parse('C10H10')
        })
        self.assertEqual(ambig, ambig if ambig in all_perm else set())

    def test_predict_compound_pairs_unbalanced(self):
        """Test prediction of (non-sense) unbalanced reaction."""
        reaction = parse_reaction(
            'a[c] <=> b[c] + c[c]')
        formulas = {
            'a': Formula.parse('C10H12'),
            'b': Formula.parse('C9H11'),
            'c': Formula.parse('CO2')
        }
        transfer, balance, ambiguous_pairs = \
            findprimarypairs.predict_compound_pairs(
                reaction, formulas)

        self.assertEqual(balance, {
            (Compound('a', 'c'), 1): Formula.parse('H'),
            (Compound('c', 'c'), 1): Formula.parse('O2'),
        })
        self.assertEqual(transfer, {
            ((Compound('a', 'c'), 1), (Compound('b', 'c'), 1)):
                Formula.parse('C9H11'),
            ((Compound('a', 'c'), 1), (Compound('c', 'c'), 1)):
                Formula.parse('C'),
        })
        self.assertEqual(ambiguous_pairs, set())

    def test_predict_compound_pairs_multiple(self):
        """Test prediction of reaction with multiple instances."""
        reaction = parse_reaction(
            'a[c] <=> (2) b[c] + c[c]')
        formulas = {
            'a': Formula.parse('C10H13O6'),
            'b': Formula.parse('C5H6O3'),
            'c': Formula.parse('H'),
        }
        transfer, balance, ambiguous_pairs = \
            findprimarypairs.predict_compound_pairs(
                reaction, formulas)

        self.assertEqual(balance, {})
        self.assertEqual(transfer, {
            ((Compound('a', 'c'), 1), (Compound('b', 'c'), 1)):
                Formula.parse('C5H6O3'),
            ((Compound('a', 'c'), 1), (Compound('b', 'c'), 2)):
                Formula.parse('C5H6O3'),
            ((Compound('a', 'c'), 1), (Compound('c', 'c'), 1)):
                Formula.parse('H')
        })
        self.assertEqual(ambiguous_pairs, set())



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2015-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

from __future__ import unicode_literals

import logging
import random

from ..command import Command, SolverCommandMixin, \
    FilePrefixAppendAction, convert_to_unicode
from ..formula import Formula, Atom, Radical, ParseError
from .. import findprimarypairs
from .. import mapmaker

from six import text_type, iteritems

logger = logging.getLogger(__name__)


class PrimaryPairsCommand(SolverCommandMixin, Command):
    """Predict primary pairs of reactions.

    This command is used to predict element-transferring reactant/product pairs
    in the reactions of the model. This can be used to determine the flow of
    elements through reactions.
    """

    @classmethod
    def init_parser(cls, parser):
        parser.add_argument(
            '--method',
            choices=['fpp', 'mapmaker'],
            default='fpp', help='Primary pair prediction method')
        parser.add_argument(
            '--exclude', metavar='reaction', type=convert_to_unicode,
            default=[], action=FilePrefixAppendAction,
            help=('Reaction to exclude (e.g. biomass reactions or'
                  ' macromolecule synthesis)'))
        parser.add_argument(
            '--report-element', metavar='element', type=text_type, default=[],
            action='append', help=('Only output pairs predicted to transfer'
                                   ' this element (e.g. C, N, P)'))
        parser.add_argument(
            '--report-all-transfers', action='store_true',
            help=('Report all transfers instead of combining elements for'
                  ' each pair (only available for fpp).'))
        parser.add_argument(
            '--ambiguous', action='store_true',
            help=('Report additional information about reactions where'
                  'primary pair identification was ambiguous'))
        parser.add_argument(
            '--weights', action='append', default=[], type=text_type,
            help=('Set weights for elements for inferring compound'
                  ' similarities (e.g. --weights N=0.4,C=1,H=0,R=20,*=0.6).'
                  ' The default value depends on the prediction method.'))
        super(PrimaryPairsCommand, cls).init_parser(parser)

    def run(self):
        # Check that elements are valid
        elements = set()
        for element in self._args.report_element:
            if not hasattr(Atom, element):
                self.argument_error('Invalid element: {}'.format(element))
            elements.add(Atom(element))

        # Check that method can report all transfers if requested
        if self._args.report_all_transfers and self._args.method != 'fpp':
            self.argument_error(
                'Reporting all transfers is not available for this method:'
                ' {}'.format(self._args.method))

        if len(self._args.weights) > 0:
            weights_dict, r_group_weight, default_weight = _parse_weights(
                self._args.weights, default_weight=0.6)

            def element_weight(element):
                if isinstance(element, Radical):
                    return r_group_weight
                return weights_dict.get(element, default_weight)
        else:
            if self._args.method == 'fpp':
                logger.info(
                    'Using default element weights for {}:'
                    ' C=1, H=0, *=0.82'.format(self._args.method))
                element_weight = findprimarypairs.element_weight
            elif self._args.method == 'mapmaker':
                logger.info(
                    'Using default element weights for {}:'
                    ' H=0, N=0.4, O=0.4, P=0.4, R=40, *=1'.format(
                        self._args.method))
                element_weight = mapmaker.default_weight

        # Mapping from compound id to formula
        compound_formula = {}
        for compound in self._model.compounds:
            if compound.formula is not None:
                try:
                    f = Formula.parse(compound.formula).flattened()
                    if not f.is_variable():
                        compound_formula[compound.id] = f
                    else:
                        logger.warning(
                            'Skipping variable formula {}: {}'.format(
                                compound.id, compound.formula))
                except ParseError as e:
                    msg = (
                        'Error parsing formula'
                        ' for compound {}:\n{}\n{}'.format(
                            compound.id, e, compound.formula))
                    if e.indicator is not None:
                        msg += '\n{}'.format(e.indicator)
                    logger.warning(msg)

        # Set of excluded reactions
        exclude = set(self._args.exclude)

        def iter_reactions():
            for reaction in self._model.reactions:
                if (reaction.id not in self._model.model or
                        reaction.id in exclude):
                    continue

                if reaction.equation is None:
                    logger.warning(
                        'Reaction {} has no reaction equation'.format(
                            reaction.id))
                    continue

                if any(c.name not in compound_formula
                       for c, _ in reaction.equation.compounds):
                    logger.warning(
                        'Reaction {} has compounds with undefined'
                        ' formula'.format(reaction.id))
                    continue

                yield reaction

        if self._args.method == 'fpp':
            result = self._run_find_primary_pairs(
                compound_formula, iter_reactions(), element_weight)
        elif self._args.method == 'mapmaker':
            result = self._run_mapmaker(
                compound_formula, iter_reactions(), element_weight)
        else:
            self.argument_error('Unknown method: {}'.format(self._args.method))

        if not self._args.report_all_transfers:
            result = self._combine_transfers(result)

        for reaction_id, c1, c2, form in result:
            if len(elements) == 0 or any(e in form for e in elements):
                print('{}\t{}\t{}\t{}'.format(reaction_id, c1, c2, form))

    def _combine_transfers(self, result):
        """Combine multiple pair transfers into one."""
        transfers = {}
        for reaction_id, c1, c2, form in result:
            key = reaction_id, c1, c2
            combined_form = transfers.setdefault(key, Formula())
            transfers[key] = combined_form | form

        for (reaction_id, c1, c2), form in iteritems(transfers):
            yield reaction_id, c1, c2, form

    def _run_find_primary_pairs(
            self, compound_formula, reactions, element_weight):
        reaction_pairs = [(r.id, r.equation) for r in reactions]
        ambig = self._args.ambiguous
        prediction, _ = findprimarypairs.predict_compound_pairs_iterated(
            reaction_pairs, compound_formula,
            ambig, element_weight=element_weight)

        for reaction_id, _ in reaction_pairs:
            if reaction_id not in prediction:
                logger.warning('Failed to predict pairs for {}'.format(
                    reaction_id))
                continue

            pairs, balance = prediction[reaction_id]
            if len(balance) > 0:
                logger.warning('Reaction {} is not balanced!'.format(
                    reaction_id))

            for (c1, c2), forms in sorted(iteritems(pairs)):
                for form in forms:
                    yield reaction_id, c1, c2, form

    def _run_mapmaker(self, compound_formula, reactions, element_weight):
        solver = self._get_solver(integer=True)
        ambiguous = set()
        for reaction in reactions:
            logger.info('Predicting reaction {}...'.format(reaction.id))
            try:
                transfer = list(mapmaker.predict_compound_pairs(
                                reaction.equation, compound_formula, solver,
                                weight_func=element_weight))
                if self._args.ambiguous:
                    if len(transfer) > 1:
                        ambiguous.add(reaction.id)
            except mapmaker.UnbalancedReactionError:
                logger.info('Reaction {} is not balanced! Skipping...'.format(
                    reaction.id))
                continue
            if len(transfer) == 0:
                logger.warning('Reaction {} has no predicted transfer!'.format(
                    reaction.id))
                continue

            for (c1, c2), form in sorted(
                    iteritems(random.sample(transfer, 1)[0])):
                yield reaction.id, c1, c2, form
        if self._args.ambiguous:
            for rxn_ambig in ambiguous:
                logger.info('Ambiguous Transfers '
                            'in Reaction: {}'.format(rxn_ambig))


def _parse_weights(weight_args, default_weight=0.6):
    """Parse list of weight assignments."""
    weights_dict = {}
    r_group_weight = default_weight
    for weight_arg in weight_args:
        for weight_assignment in weight_arg.split(','):
            if '=' not in weight_assignment:
                raise ValueError(
                    'Invalid weight assignment: {}'.format(weight_assignment))

            key, value = weight_assignment.split('=', 1)
            value = float(value)
            if key == 'R':
                r_group_weight = value
            elif key == '*':
                default_weight = value
            elif hasattr(Atom, key):
                weights_dict[Atom(key)] = value
            else:
                raise ValueError('Invalid element: {}'.format(key))

    return weights_dict, r_group_weight, default_weight



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@uri.edu>
# Copyright 2020-2020  Elysha Sameth <esameth1@my.uri.edu>


from __future__ import unicode_literals

from psamm import graph
import os
import tempfile
import unittest
from collections import defaultdict
from psamm.datasource.reaction import parse_reaction, parse_compound, \
    Direction, Compound
from psamm.datasource.native import NativeModel, ReactionEntry, CompoundEntry
from psamm.formula import Formula, Atom, ParseError


class TestGraph(unittest.TestCase):
    def setUp(self):
        self.fum = CompoundEntry({
            'id': 'fum_c', 'formula': 'C4H2O4'})
        self.rxn1 = ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum_c[c] + h2o_c[c] <=> mal_L_c[c]')})
        self.rxn2 = ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction(
                'fum_c[c] + h2o_c[c] <=> mal_L_c[c]')})
        self.g = graph.Graph()
        self.node1 = graph.Node({'id': 'A'})
        self.node2 = graph.Node({'id': 'B', 'color': 'blue'})
        self.node3 = graph.Node({'id': 'C', 'color': 'red'})
        self.node4 = graph.Node({'id': 'D',
                                 'original_id': ['A', 'B'], 'type': 'rxn'})
        self.node5 = graph.Node({'id': 'E',
                                 'original_id': 'cpd_E', 'type': 'cpd'})
        self.node6 = graph.Node({'id': 'Ex_e',
                                 'type': 'Ex_rxn'})
        self.node7 = graph.Node({'id': 'Ex_e',
                                 'type': 'cpd', 'entry': [self.fum]})
        self.node8 = graph.Node({'id': 'rxn1',
                                 'type': 'cpd', 'entry': [self.rxn1]})
        self.node9 = graph.Node({'id': 'rxn1_rxn2',
                                 'type': 'cpd', 'entry':
                                     [self.rxn1, self.rxn2]})
        self.edge1_2 = graph.Edge(self.node1, self.node2)
        self.edge2_3 = graph.Edge(self.node2, self.node3,
                                  props={'id': '2_3'})

    def test_add_node(self):
        self.g.add_node(self.node1)
        self.assertTrue(self.node1 in self.g.nodes)

    def test_node_id_dict(self):
        self.g.add_node(self.node1)
        self.assertEqual(self.node1, self.g._nodes_id[self.node1.props['id']])

    def test_add_multiple_nodes(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.assertTrue(all(i in self.g.nodes
                            for i in [self.node1, self.node2]))

    def test_original_id_EX(self):
        self.g.add_node(self.node6)
        nd = defaultdict(list)
        nd['Ex_e'].append(self.node6)
        self.assertTrue(self.g._nodes_original_id == nd)

    def test_original_id_cpd(self):
        self.g.add_node(self.node7)
        nd = defaultdict(list)
        nd[self.fum.id].append(self.node7)
        self.assertTrue(self.g._nodes_original_id == nd)

    def test_original_id_rxn(self):
        self.g.add_node(self.node8)
        nd = defaultdict(list)
        nd[self.rxn1.id].append(self.node8)
        self.assertTrue(self.g._nodes_original_id == nd)

    def test_original_id_rxn(self):
        self.g.add_node(self.node9)
        nd = defaultdict(list)
        nd['rxn1,rxn2'].append(self.node9)
        self.assertEqual(self.g._nodes_original_id, nd)

    def test_node_count(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.assertEqual(self.g.node_count, 3)

    def test_add_edge(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.g.add_edge(self.edge1_2)
        self.assertTrue(self.edge1_2 in self.g.edges)

    def test_add_edge_with_props(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.assertTrue(self.edge2_3 in self.g.edges)

    def test_add_multiple_edges(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge1_2)
        self.g.add_edge(self.edge2_3)
        self.assertTrue(all(i in self.g.edges
                            for i in [self.edge1_2, self.edge2_3]))

    def test_edge_count(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge1_2)
        self.g.add_edge(self.edge2_3)
        self.assertTrue(self.g.edge_count, 2)

    def test_add_edge_missing_node(self):
        self.g.add_node(self.node1)
        with self.assertRaises(ValueError):
            self.g.add_edge(self.edge1_2)

    def test_edges_for(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge1_2)
        self.g.add_edge(self.edge2_3)
        edges_for_list = []
        for i in self.g.edges_for(self.node1):
            edges_for_list.append(i)
        self.assertTrue(edges_for_list == [(self.node2,
                                            set([self.edge1_2]))])
        self.assertEqual('A', 'A')

    def test_get_node(self):
        self.g.add_node(self.node1)
        test_node = self.g.get_node('A')
        self.assertEqual(test_node, self.node1)

    def test_nodes_id_dict(self):
        self.g.add_node(self.node1)
        self.g.add_node(self.node2)
        self.assertEqual(self.g.nodes_id_dict, {'A': self.node1,
                                                'B': self.node2})

    def test_default_edge_props(self):
        self.g._default_edge_props['style'] = 'dashed'
        d = {'style': 'dashed'}
        self.assertEqual(d, self.g.default_edge_props)

    def test_default_node_props(self):
        self.g._default_node_props['shape'] = 'box'
        d = {'shape': 'box'}
        self.assertEqual(d, self.g.default_node_props)

    def test_write_nodes_table(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_node')
        with open(path, mode='w') as f:
            self.g.write_nodes_tables(f)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file, ['id\tcolor\tlabel\n',
                                     'B\tblue\tB\n', 'C\tred\tC\n'])

    def test_write_nodes_table_with_original_id(self):
        self.g.add_node(self.node2)
        self.node3.props['original_id'] = 'A_1'
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_node')
        with open(path, mode='w') as f:
            self.g.write_nodes_tables(f)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file, ['id\tcolor\tlabel\n',
                                     'B\tblue\tB\n', 'C\tred\tC\n'])

    def test_write_edges_tables(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_edges_tables(f)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file, ['source\ttarget\tid\n', 'B\tC\t2_3\n'])

    def test_write_graphviz(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz(f, 2, 2)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(['digraph {\n', 'size = "2, 2"; ratio = fill;\n',
                          ' "B"[color="blue",id="B"]\n',
                          ' "C"[color="red",id="C"]\n',
                          ' "B" -> "C"[id="2_3"]\n', '}\n'], read_file)

    def test_write_graphviz_graph_props(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g.props['fontsize'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz(f, 2, 2)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(['digraph {\n', 'size = "2, 2"; ratio = fill;\n',
                          ' fontsize="12";\n', ' "B"[color="blue",id="B"]\n',
                          ' "C"[color="red",id="C"]\n',
                          ' "B" -> "C"[id="2_3"]\n', '}\n'], read_file)

    def test_write_graphviz_default_node_props(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g._default_node_props['fontname'] = 'Arial'
        self.g._default_node_props['fontsize'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz(f, 2, 2)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(['digraph {\n', 'size = "2, 2"; ratio = fill;\n',
                          ' node[fontname="Arial",fontsize="12"];\n',
                          ' "B"[color="blue",id="B"]\n',
                          ' "C"[color="red",id="C"]\n',
                          ' "B" -> "C"[id="2_3"]\n', '}\n'], read_file)

    def test_write_graphvizdefault_edge_props(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g._default_edge_props['style'] = 'dashed'
        self.g._default_edge_props['width'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz(f, 2, 2)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(['digraph {\n', 'size = "2, 2"; ratio = fill;\n',
                          ' edge[style="dashed",width="12"];\n',
                          ' "B"[color="blue",id="B"]\n',
                          ' "C"[color="red",id="C"]\n',
                          ' "B" -> "C"[id="2_3"]\n', '}\n'], read_file)

    def test_write_graphviz_default_size(self):
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g._default_node_props['fontname'] = 'Arial'
        self.g._default_node_props['fontsize'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz(f, None, None)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n',
                          ' node[fontname="Arial",fontsize="12"];\n',
                          ' "B"[color="blue",id="B"]\n',
                          ' "C"[color="red",id="C"]\n',
                          ' "B" -> "C"[id="2_3"]\n', '}\n'])

    def test_write_graphviz_compartmentalize(self):
        self.node2.props['compartment'] = 'c'
        self.node3.props['compartment'] = 'e'
        self.edge2_3.props['compartment'] = 'e'
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz_compartmentalized(
                f, {'e': {'c'}, 'c': set()}, 'e', 2, 2)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n',
                          'size="2,2"; ratio = fill;\n',
                          ' subgraph cluster_e {\n',
                          '  style=solid;\n', '  color=black;\n',
                          '  penwidth=4;\n', '  fontsize=35;\n',
                          '  label = "Compartment: e"\n',
                          ' "C"[color="red",compartment="e",id="C"]\n',
                          ' subgraph cluster_c {\n', '  style=dashed;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n', '  label = "Compartment: c"\n',
                          ' "B"[color="blue",compartment="c",id="B"]\n',
                          '}} "B" -> "C"[compartment="e",id="2_3"]\n', '}\n'])

    def test_write_graphviz_compartmentalize_default_size(self):
        self.node2.props['compartment'] = 'c'
        self.node3.props['compartment'] = 'e'
        self.edge2_3.props['compartment'] = 'e'
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz_compartmentalized(
                f, {'e': {'c'}, 'c': set()}, 'e', None, None)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n', ' subgraph cluster_e {\n',
                          '  style=solid;\n', '  color=black;\n',
                          '  penwidth=4;\n', '  fontsize=35;\n',
                          '  label = "Compartment: e"\n',
                          ' "C"[color="red",compartment="e",id="C"]\n',
                          ' subgraph cluster_c {\n', '  style=dashed;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n',
                          '  label = "Compartment: c"\n',
                          ' "B"[color="blue",compartment="c",id="B"]\n',
                          '}} "B" -> "C"[compartment="e",id="2_3"]\n', '}\n'])

    def test_write_graphviz_compartmentalize_default_node_props(self):
        self.node2.props['compartment'] = 'c'
        self.node3.props['compartment'] = 'e'
        self.edge2_3.props['compartment'] = 'e'
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g._default_node_props['fontname'] = 'Arial'
        self.g._default_node_props['fontsize'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz_compartmentalized(
                f, {'e': {'c'}, 'c': set()}, 'e', None, None)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n',
                          ' node[fontname="Arial",fontsize="12"];\n',
                          ' subgraph cluster_e {\n', '  style=solid;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n', '  label = "Compartment: e"\n',
                          ' "C"[color="red",compartment="e",id="C"]\n',
                          ' subgraph cluster_c {\n', '  style=dashed;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n', '  label = "Compartment: c"\n',
                          ' "B"[color="blue",compartment="c",id="B"]\n',
                          '}} "B" -> "C"[compartment="e",id="2_3"]\n', '}\n'])

    def test_write_graphviz_compartmentalize_default_edge_props(self):
        self.node2.props['compartment'] = 'c'
        self.node3.props['compartment'] = 'e'
        self.edge2_3.props['compartment'] = 'e'
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g._default_edge_props['style'] = 'dashed'
        self.g._default_edge_props['width'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz_compartmentalized(
                f, {'e': {'c'}, 'c': set()}, 'e', None, None)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n',
                          ' edge[style="dashed",width="12"];\n',
                          ' subgraph cluster_e {\n', '  style=solid;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n', '  label = "Compartment: e"\n',
                          ' "C"[color="red",compartment="e",id="C"]\n',
                          ' subgraph cluster_c {\n', '  style=dashed;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n', '  label = "Compartment: c"\n',
                          ' "B"[color="blue",compartment="c",id="B"]\n',
                          '}} "B" -> "C"[compartment="e",id="2_3"]\n', '}\n'])

    def test_write_graphviz_compartmentalize_default_graph_props(self):
        self.node2.props['compartment'] = 'c'
        self.node3.props['compartment'] = 'e'
        self.edge2_3.props['compartment'] = 'e'
        self.g.add_node(self.node2)
        self.g.add_node(self.node3)
        self.g.add_edge(self.edge2_3)
        self.g.props['fontsize'] = 12
        path = os.path.join(tempfile.mkdtemp(), 'tmp_cyto_edge')
        with open(path, mode='w') as f:
            self.g.write_graphviz_compartmentalized(
                f, {'e': {'c'}, 'c': set()}, 'e', None, None)
        read_file = []
        f = open(path, 'r')
        for i in f.readlines():
            read_file.append(i)
        f.close()
        self.assertEqual(read_file,
                         ['digraph {\n', ' fontsize="12";\n',
                          ' subgraph cluster_e {\n',
                          '  style=solid;\n', '  color=black;\n',
                          '  penwidth=4;\n', '  fontsize=35;\n',
                          '  label = "Compartment: e"\n',
                          ' "C"[color="red",compartment="e",id="C"]\n',
                          ' subgraph cluster_c {\n', '  style=dashed;\n',
                          '  color=black;\n', '  penwidth=4;\n',
                          '  fontsize=35;\n',
                          '  label = "Compartment: c"\n',
                          ' "B"[color="blue",compartment="c",id="B"]\n',
                          '}} "B" -> "C"[compartment="e",id="2_3"]\n', '}\n'])


class TestNodes(unittest.TestCase):
    def setUp(self):
        self.g = graph.Graph()
        self.node1 = graph.Node({'id': 'A'})
        self.node2 = graph.Node({'id': 'B'})
        self.node3 = graph.Node({'id': 'C'})
        self.edge1_2 = graph.Edge(self.node1, self.node2)

    def test_eq_nodes(self):
        self.g.add_node(self.node1)
        self.assertEqual(self.node1 == self.g.get_node('A'),
                         self.node1.props == self.g.get_node('A').props)

    def test_neq_nodes(self):
        self.assertTrue(self.node1 != self.node2)

    def test_neq_node_edge(self):
        self.assertFalse(self.node1 == self.edge1_2)

    def test_node_repr(self):
        rep = self.node1.__repr__()
        self.assertEqual(rep, '<Node id=A>')


class TestOther(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        self.rxn1 = ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum_c[c] + h2o_c[c] <=> mal_L_c[c]')})
        native_model.reactions.add_entry(self.rxn1)

        self.fum = CompoundEntry({
            'id': 'fum_c', 'formula': 'C4H2O4'})
        native_model.compounds.add_entry(self.fum)

        self.h2o = CompoundEntry({
            'id': 'h2o_c', 'formula': 'H2O'})
        native_model.compounds.add_entry(self.h2o)

        self.mal = CompoundEntry(
            {'id': 'mal_L_c', 'formula': 'C4H4O5'})
        native_model.compounds.add_entry(self.mal)

        self.rxn2 = ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction(
                'q8_c[c] + succ_c[c] => fum_c[c] + q8h2_c[c]')})
        native_model.reactions.add_entry(self.rxn2)

        self.q8 = CompoundEntry(
            {'id': 'q8_c', 'formula': 'C49H74O4'})
        native_model.compounds.add_entry(self.q8)

        self.q8h2 = CompoundEntry({
            'id': 'q8h2_c', 'formula': 'C49H76O4'})
        native_model.compounds.add_entry(self.q8h2)

        self.succ = CompoundEntry({
            'id': 'succ_c', 'formula': 'C4H4O4'})
        native_model.compounds.add_entry(self.succ)

        self.native = native_model
        self.mm = native_model.create_metabolic_model()

    def test_compound_dict(self):
        cpd_dict = graph.get_compound_dict(self.native)
        test_dict = {'fum_c': Formula({Atom('O'): 4,
                                       Atom('C'): 4, Atom('H'): 2}),
                     'h2o_c': Formula({Atom('O'): 1,
                                       Atom('H'): 2}),
                     'mal_L_c': Formula({Atom('O'): 5,
                                         Atom('C'): 4, Atom('H'): 4}),
                     'q8_c': Formula({Atom('O'): 4,
                                      Atom('C'): 49, Atom('H'): 74}),
                     'q8h2_c': Formula({Atom('O'): 4,
                                        Atom('C'): 49, Atom('H'): 76}),
                     'succ_c': Formula({Atom('O'): 4,
                                        Atom('C'): 4, Atom('H'): 4})}
        self.assertEqual(test_dict, cpd_dict)

    def test_network_dict(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[])

        test_dict = {self.rxn1: ([(Compound(u'fum_c', u'c'),
                                   Compound(u'mal_L_c', u'c')),
                                  (Compound(u'h2o_c', u'c'),
                                   Compound(u'mal_L_c', u'c'))],
                                 Direction.Both),
                     self.rxn2: ([(Compound(u'q8_c', u'c'),
                                   Compound(u'q8h2_c', u'c')),
                                  (Compound(u'succ_c', u'c'),
                                   Compound(u'fum_c', u'c')),
                                  (Compound(u'succ_c', u'c'),
                                   Compound(u'q8h2_c', u'c'))],
                                 Direction.Right)}
        self.assertEqual(net_dict, test_dict)

    def test_network_dict_nofpp(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native, self.mm,
                                    subset=None, method='no-fpp',
                                    element=None, excluded_reactions=[])
        test_dict = {self.rxn1: ([(Compound(u'fum_c', u'c'),
                                   Compound(u'mal_L_c', u'c')),
                                  (Compound(u'h2o_c', u'c'),
                                   Compound(u'mal_L_c', u'c'))],
                                 Direction.Both),
                     self.rxn2: ([(Compound(u'q8_c', u'c'),
                                   Compound(u'fum_c', u'c')),
                                  (Compound(u'q8_c', u'c'),
                                   Compound(u'q8h2_c', u'c')),
                                  (Compound(u'succ_c', u'c'),
                                   Compound(u'fum_c', u'c')),
                                  (Compound(u'succ_c', u'c'),
                                   Compound(u'q8h2_c', u'c')),
                                  ],
                                 Direction.Right)}
        self.assertEqual(net_dict, test_dict)

    def test_reaction_dir_for(self):
        self.assertTrue(graph.dir_value(Direction.Forward) == 'forward')

    def test_reaction_dir_rev(self):
        self.assertTrue(graph.dir_value(Direction.Reverse) == 'back')

    def test_reaction_dir_both(self):
        self.assertTrue(graph.dir_value(Direction.Both) == 'both')


class TestMakeNetworks(unittest.TestCase):
    def setUp(self):
        self.native_model = NativeModel()
        self.atp = CompoundEntry({
            'id': 'atp', 'formula': 'C10H16N5O13P3'})
        self.adp = CompoundEntry({
            'id': 'adp', 'formula': 'C10H15N5O10P2'})
        self.glc = CompoundEntry({
            'id': 'glc', 'formula': 'C6H12O6'})
        self.g6p = CompoundEntry({
            'id': 'g6p', 'formula': 'C6H13O9P'})
        self.f6p = CompoundEntry({
            'id': 'f6p', 'formula': 'C6H13O9P'})
        self.rxn1 = ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'atp + glc => adp + g6p')})
        self.rxn2 = ReactionEntry({
            'id': 'rxn2', 'equation': parse_reaction(
                'g6p <=> f6p')})
        self.native_model.compounds.add_entry(self.atp)
        self.native_model.compounds.add_entry(self.adp)
        self.native_model.compounds.add_entry(self.g6p)
        self.native_model.compounds.add_entry(self.glc)
        self.native_model.compounds.add_entry(self.f6p)
        self.native_model.reactions.add_entry(self.rxn1)
        self.native_model.reactions.add_entry(self.rxn2)
        self.mm = self.native_model.create_metabolic_model()
        self.node_atp = graph.Node({'id': 'atp', 'entry': Compound('atp')})
        self.node_adp = graph.Node({'id': 'adp', 'entry': Compound('adp')})
        self.node_g6p = graph.Node({'id': 'g6p', 'entry': Compound('g6p')})
        self.node_glc = graph.Node({'id': 'glc', 'entry': Compound('glc')})
        self.node_f6p = graph.Node({'id': 'f6p', 'entry': Compound('f6p')})
        self.edge_1 = graph.Edge(self.node_glc, self.node_g6p,
                                 props={'id': 'g6p_glc_forward',
                                        'dir': 'forward'})
        self.edge_2 = graph.Edge(self.node_g6p, self.node_f6p,
                                 props={'id': 'f6p_g6p_both',
                                        'dir': 'both'})
        self.edge_3 = graph.Edge(self.node_atp, self.node_adp,
                                 props={'id': 'adp_atp_forward',
                                        'dir': 'forward'})
        self.edge_4 = graph.Edge(self.node_atp, self.node_adp,
                                 props={'id': 'adp_atp_forward',
                                        'dir': 'forward'})
        self.edge_5 = graph.Edge(self.node_atp, self.node_g6p,
                                 props={'id': 'atp_g6p_forward',
                                        'dir': 'forward'})
        self.edge_6 = graph.Edge(self.node_glc, self.node_adp,
                                 props={'id': 'adp_glc_forward',
                                        'dir': 'forward'})

        self.model_compound_entries = {}
        for cpd in self.native_model.compounds:
            self.model_compound_entries[cpd.id] = cpd
        self.node_atp_bip = graph.Node(
            {'id': 'atp', 'entry': [self.model_compound_entries['atp']],
             'compartment': None, 'type': 'cpd'})
        self.node_adp_bip = graph.Node(
            {'id': 'adp', 'entry': [self.model_compound_entries['adp']],
             'compartment': None, 'type': 'cpd'})
        self.node_g6p_bip = graph.Node(
            {'id': 'g6p', 'entry': [self.model_compound_entries['g6p']],
             'compartment': None, 'type': 'cpd'})
        self.node_glc_bip = graph.Node(
            {'id': 'glc', 'entry': [self.model_compound_entries['glc']],
             'compartment': None, 'type': 'cpd'})
        self.node_f6p_bip = graph.Node(
            {'id': 'f6p', 'entry': [self.model_compound_entries['f6p']],
             'compartment': None, 'type': 'cpd'})
        self.node_rxn1_1 = graph.Node({'id': 'rxn1_1', 'entry': [self.rxn1],
                                       'compartment': None, 'type': 'rxn'})
        self.node_rxn1_2 = graph.Node({'id': 'rxn1_2', 'entry': [self.rxn1],
                                       'compartment': None, 'type': 'rxn'})
        self.node_rxn2_1 = graph.Node({'id': 'rxn2_1', 'entry': [self.rxn2],
                                       'compartment': None, 'type': 'rxn'})
        self.edge_bip_1 = graph.Edge(self.node_atp_bip, self.node_rxn1_1,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_2 = graph.Edge(self.node_rxn1_1, self.node_adp_bip,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_3 = graph.Edge(self.node_atp_bip, self.node_rxn1_1,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_4 = graph.Edge(self.node_rxn1_1, self.node_g6p_bip,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_5 = graph.Edge(self.node_glc_bip, self.node_rxn1_2,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_6 = graph.Edge(self.node_rxn1_2, self.node_g6p_bip,
                                     props={'dir': 'forward',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_7 = graph.Edge(self.node_g6p_bip, self.node_rxn2_1,
                                     props={'dir': 'both',
                                            'style': 'solid', 'penwidth': 1})
        self.edge_bip_8 = graph.Edge(self.node_rxn2_1, self.node_f6p_bip,
                                     props={'dir': 'both',
                                            'style': 'solid', 'penwidth': 1})
        self.node_list = [
            self.node_atp_bip, self.node_adp_bip, self.node_g6p_bip,
            self.node_glc_bip, self.node_f6p_bip,
            self.node_rxn1_1, self.node_rxn1_2, self.node_rxn2_1]
        self.edge_list = [
            self.edge_bip_1, self.edge_bip_2, self.edge_bip_3,
            self.edge_bip_4, self.edge_bip_5, self.edge_bip_6,
            self.edge_bip_7, self.edge_bip_8]

        self.model_compound_entries = {}
        for cpd in self.native_model.compounds:
            self.model_compound_entries[cpd.id] = cpd

    def test_compound_graph(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        compound_graph = graph.make_compound_graph(net_dict)
        self.assertTrue(all(i in compound_graph.nodes
                            for i in [self.node_atp, self.node_adp,
                                      self.node_g6p, self.node_glc,
                                      self.node_f6p]))
        self.assertTrue(
            all(i in compound_graph.edges
                for i in [self.edge_1, self.edge_2,
                          self.edge_3, self.edge_4, self.edge_5]))
        self.assertTrue(self.edge_6 not in compound_graph.edges)

    def test_compound_graph_filter(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element='C', excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        compound_graph = graph.make_compound_graph(net_dict)
        self.assertTrue(all(i in compound_graph.nodes
                            for i in [self.node_atp, self.node_adp,
                                      self.node_g6p, self.node_glc,
                                      self.node_f6p]))
        self.assertTrue(all(i in compound_graph.edges
                            for i in [self.edge_1, self.edge_2,
                                      self.edge_3, self.edge_4]))
        self.assertTrue(self.edge_5 not in compound_graph.edges)
        self.assertTrue(self.edge_6 not in compound_graph.edges)

    def test_compound_graph_subset(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=['rxn1'], method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        compound_graph = graph.make_compound_graph(net_dict)
        self.assertTrue(
            all(i in compound_graph.nodes
                for i in [self.node_atp, self.node_adp,
                          self.node_g6p, self.node_glc]))
        self.assertTrue(self.node_f6p not in compound_graph.nodes)
        self.assertTrue(all(i in compound_graph.edges
                            for i in [self.edge_1, self.edge_3,
                                      self.edge_4, self.edge_5]))
        self.assertTrue(self.edge_2 not in compound_graph.edges)
        self.assertTrue(self.edge_6 not in compound_graph.edges)

    def test_compound_graph_exclude(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=['rxn2'],
                                    reaction_dict={}, analysis=None)
        compound_graph = graph.make_compound_graph(net_dict)
        self.assertTrue(
            all(i in compound_graph.nodes
                for i in [self.node_atp, self.node_adp,
                          self.node_g6p, self.node_glc]))
        self.assertTrue(self.node_f6p not in compound_graph.nodes)
        self.assertTrue(all(i in compound_graph.edges
                            for i in [self.edge_1, self.edge_3,
                                      self.edge_4, self.edge_5]))
        self.assertTrue(self.edge_2 not in compound_graph.edges)
        self.assertTrue(self.edge_6 not in compound_graph.edges)

    def test_compound_graph_nofpp(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='no-fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        compound_graph = graph.make_compound_graph(net_dict)
        self.assertTrue(all(i in compound_graph.nodes for i in
                            [self.node_f6p, self.node_g6p, self.node_glc,
                             self.node_adp, self.node_atp]))
        self.assertTrue(all(i in compound_graph.edges for i in
                            [self.edge_1, self.edge_2, self.edge_3,
                             self.edge_4, self.edge_5, self.edge_6]))

    def test_make_cpair_dict_combine_0(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_2']
        test_dict[(Compound('g6p'), Compound('f6p'))]['both'] = ['rxn2_1']

        test_new_id = {u'rxn2_1': self.rxn2,
                       u'rxn1_2': self.rxn1, u'rxn1_1': self.rxn1}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_make_cpair_dict_combine_1(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 1, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('g6p'), Compound('f6p'))]['both'] = ['rxn2_1']

        test_new_id = {u'rxn2_1': self.rxn2, u'rxn1_1': self.rxn1}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_make_cpair_dict_combine_2(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 2, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('g6p'), Compound('f6p'))]['both'] = ['rxn2_1']

        test_new_id = {u'rxn2_1': self.rxn2, u'rxn1_1': self.rxn1}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_make_cpair_dict_nofpp(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='no-fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'no-fpp', 0, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1']
        test_dict[(Compound('glc'), Compound('adp'))]['forward'] = ['rxn1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1']
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1']
        test_dict[(Compound('g6p'), Compound('f6p'))]['both'] = ['rxn2']

        test_new_id = {u'rxn2': self.rxn2, u'rxn1': self.rxn1}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_cpair_dict_duplicate_cpair_both(self):
        rxn3 = ReactionEntry({
            'id': 'rxn3', 'equation': parse_reaction(
                'f6p <=> g6p')})
        nm = self.native_model
        nm.reactions.add_entry(rxn3)
        mm = self.native_model.create_metabolic_model()

        net_dict, style_flux_dict = \
            graph.make_network_dict(nm, mm, subset=None,
                                    method='fpp', element=None,
                                    excluded_reactions=[], reaction_dict={},
                                    analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_2']
        test_dict[(Compound('f6p'),
                   Compound('g6p'))]['both'] = ['rxn2_1', 'rxn3_1']

        test_new_id = {u'rxn2_1': self.rxn2, u'rxn1_2': self.rxn1,
                       u'rxn1_1': self.rxn1, 'rxn3_1': rxn3}
        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_cpair_dict_duplicate_cpair_rev(self):
        rxn3 = ReactionEntry({
            'id': 'rxn3', 'equation': parse_reaction(
                'g6p => f6p')})
        nm = self.native_model
        nm.reactions.add_entry(rxn3)
        mm = self.native_model.create_metabolic_model()

        net_dict, style_flux_dict = \
            graph.make_network_dict(nm, mm, subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_2']
        test_dict[(Compound('g6p'), Compound('f6p'))]['both'] = ['rxn2_1']
        test_dict[(Compound('g6p'), Compound('f6p'))]['forward'] = ['rxn3_1']

        test_new_id = {u'rxn1_1': self.rxn1, u'rxn1_2': self.rxn1,
                       u'rxn2_1': self.rxn2, 'rxn3_1': rxn3}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_cpair_dict_duplicate_for(self):
        rxn3 = ReactionEntry({
            'id': 'rxn3', 'equation': parse_reaction(
                'f6p => g6p')})
        nm = self.native_model
        nm.reactions.add_entry(rxn3)
        mm = self.native_model.create_metabolic_model()

        net_dict, style_flux_dict = \
            graph.make_network_dict(nm, mm, subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        test_dict = defaultdict(lambda: defaultdict(list))
        test_dict[(Compound('atp'), Compound('adp'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('atp'), Compound('g6p'))]['forward'] = ['rxn1_1']
        test_dict[(Compound('glc'), Compound('g6p'))]['forward'] = ['rxn1_2']
        test_dict[(Compound('f6p'), Compound('g6p'))]['both'] = ['rxn2_1']
        test_dict[(Compound('f6p'), Compound('g6p'))]['forward'] = ['rxn3_1']

        test_new_id = {u'rxn2_1': self.rxn2, u'rxn1_2': self.rxn1,
                       u'rxn1_1': self.rxn1, 'rxn3_1': rxn3}

        self.assertEqual(cpairs, test_dict)
        self.assertEqual(new_id, test_new_id)

    def test_node_original_id_dict(self):
        g = graph.Graph()
        g.add_node(self.node_atp)
        d = defaultdict(list)
        d['atp'].append(self.node_atp)
        self.assertEqual(g.nodes_original_id_dict, d)

    def test_bipartite_graph(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        self.assertTrue(all(i in bipartite_graph.nodes
                            for i in self.node_list))
        self.assertTrue(all(i in self.node_list
                            for i in bipartite_graph.nodes))
        self.assertTrue(all(i in bipartite_graph.edges
                            for i in self.edge_list))
        self.assertTrue(all(i in self.edge_list
                            for i in bipartite_graph.edges))

    def test_bipartite_graph_filter(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element='C', excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        edge_list = [self.edge_bip_1, self.edge_bip_2, self.edge_bip_5,
                     self.edge_bip_6, self.edge_bip_7, self.edge_bip_8]
        self.assertTrue(all(i in bipartite_graph.nodes
                            for i in self.node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in self.node_list
                            for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_filter2(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element='P', excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        node_list = [i for i in self.node_list
                     if i not in [self.node_glc_bip, self.node_rxn1_2]]
        edge_list = [self.edge_bip_1, self.edge_bip_2, self.edge_bip_3,
                     self.edge_bip_4, self.edge_bip_7, self.edge_bip_8]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_filter3(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        edge_list = [self.edge_bip_1, self.edge_bip_2, self.edge_bip_3,
                     self.edge_bip_4, self.edge_bip_5, self.edge_bip_6,
                     self.edge_bip_7, self.edge_bip_8]
        self.assertTrue(all(i in bipartite_graph.nodes
                            for i in self.node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in self.node_list
                            for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_subset(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=['rxn1'], method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        node_list = [i for i in self.node_list
                     if i not in [self.node_f6p_bip, self.node_rxn2_1]]
        edge_list = [self.edge_bip_1, self.edge_bip_2, self.edge_bip_3,
                     self.edge_bip_4, self.edge_bip_5, self.edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_subset2(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=['rxn2'], method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        node_list = [self.node_g6p_bip, self.node_f6p_bip, self.node_rxn2_1]
        edge_list = [self.edge_bip_7, self.edge_bip_8]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_exclude(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=['rxn2'],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        node_list = [i for i in self.node_list
                     if i not in [self.node_f6p_bip, self.node_rxn2_1]]
        edge_list = [self.edge_bip_1, self.edge_bip_2, self.edge_bip_3,
                     self.edge_bip_4, self.edge_bip_5, self.edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_nofpp(self):
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='no-fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict={}, analysis=None)
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'no-fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'no-fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict)
        node_rxn1 = graph.Node({'id': 'rxn1', 'entry': [self.rxn1],
                                'compartment': None, 'type': 'rxn'})
        node_rxn2 = graph.Node({'id': 'rxn2', 'entry': [self.rxn2],
                                'compartment': None, 'type': 'rxn'})
        edge_bip_1 = graph.Edge(self.node_atp_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 1})
        edge_bip_2 = graph.Edge(self.node_glc_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 1})
        edge_bip_3 = graph.Edge(node_rxn1, self.node_adp_bip,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 1})
        edge_bip_4 = graph.Edge(node_rxn1, self.node_g6p_bip,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 1})
        edge_bip_5 = graph.Edge(self.node_g6p_bip, node_rxn2,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 1})
        edge_bip_6 = graph.Edge(node_rxn2, self.node_f6p_bip,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 1})
        node_list = [self.node_atp_bip, self.node_adp_bip, self.node_g6p_bip,
                     self.node_glc_bip, self.node_f6p_bip,
                     node_rxn1, node_rxn2]
        edge_list = [edge_bip_1, edge_bip_2, edge_bip_3,
                     edge_bip_4, edge_bip_5, edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_fba(self):
        reaction_dict = {'rxn1': (0, 1), 'rxn2': (2, 1)}
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='no-fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict=reaction_dict,
                                    analysis='fba')
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'no-fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'no-fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict, 'fba')
        node_rxn1 = graph.Node({'id': 'rxn1', 'entry': [self.rxn1],
                                'compartment': None, 'type': 'rxn'})
        node_rxn2 = graph.Node({'id': 'rxn2', 'entry': [self.rxn2],
                                'compartment': None, 'type': 'rxn'})
        edge_bip_1 = graph.Edge(self.node_atp_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_2 = graph.Edge(self.node_glc_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_3 = graph.Edge(node_rxn1, self.node_adp_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_4 = graph.Edge(node_rxn1, self.node_g6p_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_5 = graph.Edge(self.node_g6p_bip, node_rxn2,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 5})
        edge_bip_6 = graph.Edge(node_rxn2, self.node_f6p_bip,
                                props={'dir': 'forward',
                                       'style': 'solid', 'penwidth': 5})
        node_list = [self.node_atp_bip, self.node_adp_bip,
                     self.node_g6p_bip, self.node_glc_bip,
                     self.node_f6p_bip, node_rxn1, node_rxn2]
        edge_list = [edge_bip_1, edge_bip_2, edge_bip_3,
                     edge_bip_4, edge_bip_5, edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_fva(self):
        reaction_dict = {'rxn1': (0, 0), 'rxn2': (-1, 1)}
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='no-fpp',
                                    element='C', excluded_reactions=[],
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'no-fpp', 0, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'no-fpp', 0,
                                              self.model_compound_entries,
                                              new_style_flux_dict, 'fva')
        node_rxn1 = graph.Node({'id': 'rxn1', 'entry': [self.rxn1],
                                'compartment': None, 'type': 'rxn'})
        node_rxn2 = graph.Node({'id': 'rxn2', 'entry': [self.rxn2],
                                'compartment': None, 'type': 'rxn'})
        edge_bip_1 = graph.Edge(self.node_atp_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_2 = graph.Edge(self.node_glc_bip, node_rxn1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_3 = graph.Edge(node_rxn1, self.node_adp_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_4 = graph.Edge(node_rxn1, self.node_g6p_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_5 = graph.Edge(self.node_g6p_bip, node_rxn2,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 5})
        edge_bip_6 = graph.Edge(node_rxn2, self.node_f6p_bip,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 5})
        node_list = [self.node_atp_bip, self.node_adp_bip,
                     self.node_g6p_bip, self.node_glc_bip, self.node_f6p_bip,
                     node_rxn1, node_rxn2]
        edge_list = [edge_bip_1, edge_bip_2, edge_bip_3,
                     edge_bip_4, edge_bip_5, edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))

    def test_bipartite_graph_combine_2(self):
        reaction_dict = {'rxn1': (0, 0), 'rxn2': (-1, 1)}
        net_dict, style_flux_dict = \
            graph.make_network_dict(self.native_model, self.mm,
                                    subset=None, method='fpp',
                                    element=None, excluded_reactions=[],
                                    reaction_dict=reaction_dict,
                                    analysis='fva')
        cpairs, new_id, new_style_flux_dict = graph.make_cpair_dict(
            net_dict, 'fpp', 2, style_flux_dict)
        bipartite_graph = \
            graph.make_bipartite_graph_object(cpairs, new_id, 'fpp', 2,
                                              self.model_compound_entries,
                                              new_style_flux_dict, 'fva')
        node_rxn1_1 = graph.Node({'id': 'rxn1_1', 'entry': [self.rxn1],
                                  'compartment': None, 'type': 'rxn'})
        node_rxn2_1 = graph.Node({'id': 'rxn2_1', 'entry': [self.rxn2],
                                  'compartment': None, 'type': 'rxn'})
        edge_bip_1 = graph.Edge(self.node_atp_bip, node_rxn1_1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_2 = graph.Edge(node_rxn1_1, self.node_adp_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_3 = graph.Edge(node_rxn1_1, self.node_g6p_bip,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        edge_bip_4 = graph.Edge(self.node_g6p_bip, node_rxn2_1,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 5})
        edge_bip_5 = graph.Edge(node_rxn2_1, self.node_f6p_bip,
                                props={'dir': 'both',
                                       'style': 'solid', 'penwidth': 5})
        edge_bip_6 = graph.Edge(self.node_glc_bip, node_rxn1_1,
                                props={'dir': 'forward',
                                       'style': 'dotted', 'penwidth': 1})
        node_list = [self.node_atp_bip, self.node_adp_bip,
                     self.node_g6p_bip, self.node_glc_bip, self.node_f6p_bip,
                     node_rxn1_1, node_rxn2_1]
        edge_list = [edge_bip_1, edge_bip_2, edge_bip_3,
                     edge_bip_4, edge_bip_5, edge_bip_6]
        self.assertTrue(all(i in bipartite_graph.nodes for i in node_list))
        self.assertTrue(all(i in bipartite_graph.edges for i in edge_list))
        self.assertTrue(all(i in node_list for i in bipartite_graph.nodes))
        self.assertTrue(all(i in edge_list for i in bipartite_graph.edges))


class TestEdges(unittest.TestCase):
    def setUp(self):
        self.g = graph.Graph()
        self.node1 = graph.Node({'id': 'A'})
        self.node2 = graph.Node({'id': 'B'})
        self.node3 = graph.Node({'id': 'C'})
        self.node4 = graph.Node({'id': 'A'})
        self.edge1_2 = graph.Edge(self.node1, self.node2)
        self.edge2_3 = graph.Edge(self.node2, self.node3)
        self.edge12_21 = graph.Edge(self.node1, self.node2)

    def test_eq_edges(self):
        self.assertTrue(self.edge1_2 == self.edge12_21)

    def test_neq_edges(self):
        self.assertTrue(self.edge1_2 != self.edge2_3)

    def test_neq_edge_node(self):
        self.assertFalse(self.edge1_2 == self.node1)


class TestCompExit(unittest.TestCase):
    def setUp(self):
        native_model = NativeModel()
        self.rxn1 = ReactionEntry({
            'id': 'rxn1', 'equation': parse_reaction(
                'fum_c[c] + h2o_c[c] <=> mal_L_c[c]')})
        native_model.reactions.add_entry(self.rxn1)
        self.native = native_model
        self.mm = native_model.create_metabolic_model()

    def test_sys_exit(self):
        with self.assertRaises(SystemExit):
            graph.make_network_dict(self.native, self.mm)

    def test_with_element(self):
        with self.assertRaises(SystemExit):
            graph.make_network_dict(self.native, self.mm,
                                    method='no-fpp', element='C')


if __name__ == '__main__':
    unittest.main()



# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2017  Jon Lund Steffensen <jon_steffensen@uri.edu>
# Copyright 2015-2020  Keith Dufault-Thompson <keitht547@my.uri.edu>

import unittest

from psamm import mapmaker
from psamm.lpsolver import generic
from psamm.formula import Formula, Atom, Radical
from psamm.reaction import Compound
from psamm.datasource.reaction import parse_reaction


class TestMapMakerWeigths(unittest.TestCase):
    def test_default_weight(self):
        self.assertEqual(mapmaker.default_weight(Atom.C), 1)
        self.assertEqual(mapmaker.default_weight(Atom.N), 0.4)
        self.assertEqual(mapmaker.default_weight(Atom.O), 0.4)
        self.assertEqual(mapmaker.default_weight(Atom.P), 0.4)
        self.assertEqual(mapmaker.default_weight(Atom.S), 1)
        self.assertEqual(mapmaker.default_weight(Radical('R')), 40.0)

    def test_weighted_formula(self):
        weighted = mapmaker._weighted_formula(
            Formula.parse('C1H2N3S4R'), mapmaker.default_weight)
        self.assertEqual(set(weighted), {
            (Atom.C, 1, 1),
            # H is discarded
            (Atom.N, 3, 0.4),
            (Atom.S, 4, 1),
            (Radical('R'), 1, 40.0)
        })


class TestMapMaker(unittest.TestCase):
    def setUp(self):
        try:
            self.solver = generic.Solver(integer=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an MILP solver for tests')

    def test_predict_compound_pairs(self):
        """Test prediction of HEX1 reaction."""
        reaction = parse_reaction(
            'atp[c] + glc-D[c] => adp[c] + g6p[c] + h[c]')
        formulas = {
            'atp': Formula.parse('C10H12N5O13P3'),
            'adp': Formula.parse('C10H12N5O10P2'),
            'glc-D': Formula.parse('C6H12O6'),
            'g6p': Formula.parse('C6H11O9P'),
            'h': Formula.parse('H')
        }

        solutions = list(
            mapmaker.predict_compound_pairs(reaction, formulas, self.solver))

        self.assertEqual(len(solutions), 1)
        self.assertEqual(solutions[0], {
            (Compound('atp', 'c'), Compound('adp', 'c')):
                Formula.parse('C10N5O10P2'),
            (Compound('glc-D', 'c'), Compound('g6p', 'c')):
                Formula.parse('C6O6'),
            (Compound('atp', 'c'), Compound('g6p', 'c')):
                Formula.parse('O3P'),
        })

    def test_predict_compound_pairs_unbalanced(self):
        """Test prediction of (non-sense) unbalanced reaction."""
        reaction = parse_reaction(
            'a[c] <=> b[c] + c[c]')
        formulas = {
            'a': Formula.parse('C10H12'),
            'b': Formula.parse('C9H11'),
            'c': Formula.parse('CO2'),
        }

        with self.assertRaises(mapmaker.UnbalancedReactionError):
            list(mapmaker.predict_compound_pairs(
                reaction, formulas, self.solver))

    def test_predict_compound_pairs_multiple(self):
        """Test prediction of reaction with multiple instances."""
        reaction = parse_reaction(
            'a[c] <=> (2) b[c] + c[c]')
        formulas = {
            'a': Formula.parse('C10H13O6'),
            'b': Formula.parse('C5H6O3'),
            'c': Formula.parse('H'),
        }
        solutions = list(
            mapmaker.predict_compound_pairs(reaction, formulas, self.solver))

        self.assertEqual(len(solutions), 1)
        self.assertEqual(solutions[0], {
            (Compound('a', 'c'), Compound('b', 'c')): Formula.parse('C10O6'),
        })

    def test_predict_with_multiple_solutions(self):
        """Test prediction where multiple solutions exist (TALA)."""
        reaction = parse_reaction(
            'r5p[c] + xu5p-D[c] <=> g3p[c] + s7p[c]')
        formulas = {
            'r5p': Formula.parse('C5H9O8P'),
            'xu5p-D': Formula.parse('C5H9O8P'),
            'g3p': Formula.parse('C3H5O6P'),
            's7p': Formula.parse('C7H13O10P'),
        }

        solutions = list(
            mapmaker.predict_compound_pairs(reaction, formulas, self.solver))
        self.assertEqual(len(solutions), 2)

        # Solution A
        self.assertIn({
            (Compound('r5p', 'c'), Compound('s7p', 'c')):
                Formula.parse('C5O8P'),
            (Compound('xu5p-D', 'c'), Compound('g3p', 'c')):
                Formula.parse('C3O6P'),
            (Compound('xu5p-D', 'c'), Compound('s7p', 'c')):
                Formula.parse('C2O2'),
        }, solutions)

        # Solution B
        self.assertIn({
            (Compound('xu5p-D', 'c'), Compound('s7p', 'c')):
                Formula.parse('C5O8P'),
            (Compound('r5p', 'c'), Compound('g3p', 'c')):
                Formula.parse('C3O6P'),
            (Compound('r5p', 'c'), Compound('s7p', 'c')):
                Formula.parse('C2O2'),
        }, solutions)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This file is part of PSAMM.
#
# PSAMM is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PSAMM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PSAMM.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2014-2015  Jon Lund Steffensen <jon_steffensen@uri.edu>

import unittest

from psamm.metabolicmodel import MetabolicModel
from psamm.database import DictDatabase
from psamm.datasource.reaction import parse_reaction
from psamm.commands import tmfa
from psamm.lpsolver import generic
from decimal import Decimal
from psamm.lpsolver import lp


try:
    test_solver = generic.Solver()
    requires_solver = unittest.skipIf(test_solver._properties['name'] not in [
        'cplex', 'gurobi'], 'Unable to find an LP solver for tests '
                            '(TMFA requires Cplex or Gurobi as LP solver')
except generic.RequirementsError:
    requires_solver = unittest.skip('Unable to find an LP solver for tests')


@requires_solver
class TestTMFA(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('rxn_1', parse_reaction('=> (2) A[c]'))
        self.database.set_reaction('rxn_2', parse_reaction('A[c] <=> B[c]'))
        self.database.set_reaction('rxn_3', parse_reaction('A[c] => D[e]'))
        self.database.set_reaction('rxn_4', parse_reaction('A[c] => C[c]'))
        self.database.set_reaction('rxn_5', parse_reaction('C[c] => D[e]'))
        self.database.set_reaction('rxn_6', parse_reaction('D[e] =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        self.mm_irreversible, _, self.split_reversible, \
            reversible_lump_to_rxn_dict = self.model.make_irreversible(
                                                    gene_dict={},
                                                    lumped_rxns={},
                                                    exclude_list=[],
                                                    all_reversible=False)
        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        try:
            self.solver = generic.Solver(integrality_tolerance=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

    def test_mm_irrev(self):
        mm_irreversible, _, split_reversible, new_lump_rxn_dict = \
            self.model.make_irreversible({}, [], {}, False)
        self.assertTrue('rxn_2_forward' in [i for i in
                                            mm_irreversible.reactions])
        self.assertTrue('rxn_2_reverse' in [i for i in
                                            mm_irreversible.reactions])

    def test_mm_irrev_exclude(self):
        mm_irreversible, _, split_reversible, new_lump_rxn_dict = \
            self.model.make_irreversible({}, ['rxn_2'], {}, False)
        self.assertTrue('rxn_2_forward' not in [i for i in
                                                mm_irreversible.reactions])
        self.assertTrue('rxn_2_reverse' not in [i for i in
                                                mm_irreversible.reactions])

    def test_make_tmfa_problem_fluxvar(self):
        prob, v, zi, dgri, xij, cp_list = tmfa.make_tmfa_problem(
            self.mm_irreversible, self.solver)
        self.assertTrue(v.has_variable('rxn_1'))
        self.assertTrue(v.has_variable('rxn_2_forward'))
        self.assertFalse(v.has_variable('rxn_2'))

    def test_make_tmfa_problem_dgrvar(self):
        prob, v, zi, dgri, xij, cp_list = tmfa.make_tmfa_problem(
            self.mm_irreversible, self.solver)
        self.assertTrue(dgri.has_variable('rxn_1'))
        self.assertTrue(dgri.has_variable('rxn_2_forward'))
        self.assertFalse(dgri.has_variable('rxn_2'))

    def test_make_tmfa_problem_zivar(self):
        prob, v, zi, dgri, xij, cp_list = tmfa.make_tmfa_problem(
            self.mm_irreversible, self.solver)
        self.assertTrue(zi.has_variable('rxn_1'))
        self.assertTrue(zi.has_variable('rxn_2_forward'))
        self.assertFalse(zi.has_variable('rxn_2'))

    def test_make_tmfa_problem_xijvar(self):
        prob, v, zi, dgri, xij, cp_list = tmfa.make_tmfa_problem(
            self.mm_irreversible, self.solver)
        self.assertTrue(xij.has_variable('A[c]'))
        self.assertTrue(xij.has_variable('B[c]'))
        self.assertFalse(xij.has_variable('A[e]'))
        self.assertFalse(xij.has_variable('D[c]'))
        self.assertTrue(xij.has_variable('D[e]'))
        self.assertFalse(xij.has_variable('rxn1'))

    def test_make_tmfa_problem_cplist(self):
        prob, v, zi, dgri, xij, cp_list = tmfa.make_tmfa_problem(
            self.mm_irreversible, self.solver)
        self.assertEqual(set(cp_list), set(['A[c]', 'B[c]', 'C[c]', 'D[e]']))

    def test_parsedgr(self):
        f = ['rxn_1\t1\t2', 'rxn_2\t-4\t0.5']
        dgr_dict = tmfa.parse_dgr_file(f, self.mm_irreversible)
        self.assertEqual(dgr_dict, {'rxn_1': (Decimal(1), Decimal(2)),
                                    'rxn_2_forward': (Decimal(-4),
                                                      Decimal(0.5)),
                                    'rxn_2_reverse': (Decimal(4),
                                                      Decimal(0.5))})

    def test_parsedgrinvalid(self):
        f = ['rxn_1\t1\t2', 'rxn_2\tNA\tNA']
        dgr_dict = tmfa.parse_dgr_file(f, self.mm_irreversible)
        self.assertEqual(dgr_dict, {'rxn_1': (Decimal(1), Decimal(2))})


@requires_solver
class TestSolving(unittest.TestCase):
    def setUp(self):
        self.database = DictDatabase()
        self.database.set_reaction('EX_A', parse_reaction('<=> A[e]'))
        self.database.set_reaction('EX_C', parse_reaction('<=> C[e]'))
        self.database.set_reaction('EX_h2o', parse_reaction('<=> h2o[e]'))
        self.database.set_reaction('EX_h', parse_reaction('<=> h[e]'))
        self.database.set_reaction('rxn1', parse_reaction(
            'C[e] + h[e] <=> C[c] + h[c]'))
        self.database.set_reaction('rxn2', parse_reaction(
            'A[e] + h[e] <=> A[c] + h[c]'))
        self.database.set_reaction('rxn3', parse_reaction(
            'A[c] => B[c] + C[c]'))
        self.database.set_reaction('rxn4', parse_reaction('B[c] => D[c]'))
        self.database.set_reaction('rxn5', parse_reaction('D[c] => F[c]'))
        self.database.set_reaction('rxn6', parse_reaction('B[c] => E[c]'))
        self.database.set_reaction('rxn7', parse_reaction('E[c] => F[c]'))
        self.database.set_reaction('rxn8', parse_reaction('h2o[e] <=> h2o[c]'))
        self.database.set_reaction('bio', parse_reaction('F[c] =>'))
        self.model = MetabolicModel.load_model(
            self.database, self.database.reactions)
        try:
            self.solver = generic.Solver()
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        try:
            self.solver = generic.Solver(integrality_tolerance=True)
        except generic.RequirementsError:
            self.skipTest('Unable to find an LP solver for tests')

        self.dgr_dict = {'rxn1_forward': (0, 2), 'rxn1_reverse': (0, 2),
                         'rxn2_forward': (0, 2), 'rxn2_reverse': (0, 2),
                         'rxn8_forward': (0, 2), 'rxn8_reverse': (0, 2),
                         'rxn3': (-5, 2), 'rxn4': (-5, 2),
                         'rxn5': (-5, 2), 'rxn6': (300, 2), 'rxn7': (300, 2)}
        self.trans_param = {'rxn1_forward': (Decimal(1), Decimal(1)),
                            'rxn1_reverse': (-Decimal(1), -Decimal(1)),
                            'rxn2_forward': (Decimal(1), Decimal(1)),
                            'rxn2_reverse': (-Decimal(1), -Decimal(1)),
                            'rxn8_forward': (Decimal(0), Decimal(0)),
                            'rxn8_reverse': (-Decimal(0), -Decimal(0))}
        self.mm_irreversible, _, self.split_reversible, \
            self.reversible_lump_to_rxn_dict = \
            self.model.make_irreversible({},
                                         ['EX_A', 'EX_C', 'EX_h',
                                          'EX_h2o', 'bio'], {}, False)

        self.empty_prob, self.empty_v, self.empty_zi, self.empty_dgri, \
            self.empty_xij, self.empty_cp_list = tmfa.make_tmfa_problem(
                self.mm_irreversible, self.solver)
        self.prob, self.v, self.zi, self.dgri, self.xij, \
            self.cp_list = tmfa.make_tmfa_problem(
                self.mm_irreversible, self.solver)

        self.prob, self.cpd_xij_dict = tmfa.add_conc_constraints(
            self.xij, self.prob, {}, self.cp_list, ['h2o[c]', 'h2o[e]'],
            'h[c]', 'h[e]', '')
        self.prob, self.excluded_compounds = \
            tmfa.add_reaction_constraints(self.prob, self.v,
                                          self.zi, self.dgri, self.xij,
                                          self.mm_irreversible, [],
                                          ['EX_A', 'EX_C', 'EX_h',
                                           'EX_h2o', 'bio'],
                                          ['EX_A', 'EX_C', 'EX_H',
                                           'EX_h2o', 'bio'], self.dgr_dict,
                                          {}, self.split_reversible,
                                          self.trans_param,
                                          [i for i in
                                           self.mm_irreversible.reactions],
                                          None, ['h2o', 'h2o[e]'], 'h[c]',
                                          'h[e]', '', [(float(4), float(11)),
                                                       (float(4), float(11))],
                                          275, err_est=False, hamilton=False)

        self.ham_prob, self.ham_v, self.ham_zi, self.ham_dgri, \
            self.ham_xij, self.ham_cp_list = \
            tmfa.make_tmfa_problem(self.mm_irreversible, self.solver)
        self.ham_prob, self.ham_cpd_xij_dict = tmfa.add_conc_constraints(
            self.ham_xij, self.ham_prob, {}, self.cp_list,
            ['h2o[c]', 'h2o[e]'], 'h[c]', 'h[e]', '')
        self.ham_prob, self.ham_excluded_compounds = \
            tmfa.add_reaction_constraints(self.ham_prob, self.ham_v,
                                          self.ham_zi, self.ham_dgri,
                                          self.ham_xij, self.mm_irreversible,
                                          [], ['EX_A', 'EX_C', 'EX_h',
                                               'EX_h2o', 'bio'],
                                          ['EX_A', 'EX_C', 'EX_H',
                                           'EX_h2o', 'bio'], self.dgr_dict,
                                          {}, self.split_reversible,
                                          self.trans_param,
                                          [i for i in
                                           self.mm_irreversible.reactions],
                                          None, ['h2o', 'h2o[e]'], 'h[c]',
                                          'h[e]', '', [(float(4), float(11)),
                                                       (float(4), float(11))],
                                          275, err_est=False, hamilton=True)

        self.err_prob, self.err_v, self.err_zi, self.err_dgri, self.err_xij, \
            self.err_cp_list = tmfa.make_tmfa_problem(
                self.mm_irreversible, self.solver)
        self.err_prob, self.err_cpd_xij_dict = \
            tmfa.add_conc_constraints(self.err_xij, self.err_prob, {},
                                      self.cp_list, ['h2o[c]', 'h2o[e]'],
                                      'h[c]', 'h[e]', '')
        self.err_prob, self.err_excluded_compounds = \
            tmfa.add_reaction_constraints(self.err_prob, self.err_v,
                                          self.err_zi, self.err_dgri,
                                          self.err_xij, self.mm_irreversible,
                                          [], ['EX_A', 'EX_C', 'EX_h',
                                               'EX_h2o', 'bio'],
                                          ['EX_A', 'EX_C', 'EX_H',
                                           'EX_h2o', 'bio'], self.dgr_dict,
                                          {}, self.split_reversible,
                                          self.trans_param,
                                          [i for i in
                                           self.mm_irreversible.reactions],
                                          None, ['h2o', 'h2o[e]'], 'h[c]',
                                          'h[e]', '', [(float(4), float(11)),
                                                       (float(4), float(11))],
                                          275, err_est=True, hamilton=False)

    def test_addconc_constraints_default(self):
        prob, cpd_xij_dict = tmfa.add_conc_constraints(self.empty_xij,
                                                       self.empty_prob, {},
                                                       self.empty_cp_list,
                                                       ['h2o[c]', 'h2o[e]'],
                                                       'h[c]', 'h[e]', '')
        prob, excluded_compounds = \
            tmfa.add_reaction_constraints(prob, self.empty_v,
                                          self.empty_zi, self.empty_dgri,
                                          self.empty_xij,
                                          self.mm_irreversible, [],
                                          ['EX_A', 'EX_C', 'EX_h',
                                           'EX_h2o', 'bio'],
                                          ['EX_A', 'EX_C', 'EX_H',
                                           'EX_h2o', 'bio'], self.dgr_dict,
                                          {}, self.split_reversible,
                                          self.trans_param,
                                          [i for i in
                                           self.mm_irreversible.reactions],
                                          None, ['h2o', 'h2o[e]'], 'h[c]',
                                          'h[e]', '', [(float(4), float(11)),
                                                       (float(4), float(11))],
                                          275, err_est=False, hamilton=False)

        cpd_range_dict = {}
        for cpd in ['A[c]', 'B[c]', 'C[c]']:
            cpd_range_dict[cpd] = \
                (tmfa.get_var_bound(prob, self.empty_xij[cpd],
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(prob, self.empty_xij[cpd],
                                    lp.ObjectiveSense.Maximize))

        self.assertAlmostEqual(cpd_range_dict['A[c]'][0],
                               (-11.512925026140119,
                                -3.912023005428146)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['A[c]'][1],
                               (-11.512925026140119,
                                -3.912023005428146)[1], places=6)
        self.assertAlmostEqual(cpd_range_dict['B[c]'][0],
                               (-11.512925026140119,
                                -3.912023005428146)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['B[c]'][1],
                               (-11.512925026140119,
                                -3.912023005428146)[1], places=6)
        self.assertAlmostEqual(cpd_range_dict['C[c]'][0],
                               (-11.512925026140119,
                                -3.912023005428146)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['C[c]'][1],
                               (-11.512925026140119,
                                -3.912023005428146)[1], places=6)

    def test_addconc_constraints_nondefault(self):
        prob, cpd_xij_dict = \
            tmfa.add_conc_constraints(self.empty_xij,
                                      self.empty_prob,
                                      {'A[c]': (0.005, 0.005),
                                       'B[c]': (0.0005, 0.005)},
                                      self.empty_cp_list, ['h2o[c]', 'h2o[e]'],
                                      'h[c]', 'h[e]', '')
        prob, excluded_compounds = \
            tmfa.add_reaction_constraints(prob,
                                          self.empty_v, self.empty_zi,
                                          self.empty_dgri, self.empty_xij,
                                          self.mm_irreversible, [],
                                          ['EX_A', 'EX_C', 'EX_h',
                                           'EX_h2o', 'bio'],
                                          ['EX_A', 'EX_C', 'EX_H',
                                           'EX_h2o', 'bio'],
                                          self.dgr_dict, {},
                                          self.split_reversible,
                                          self.trans_param,
                                          [i for i in
                                           self.mm_irreversible.reactions],
                                          None, ['h2o', 'h2o[e]'], 'h[c]',
                                          'h[e]', '', [(float(4), float(11)),
                                                       (float(4), float(11))],
                                          275, err_est=False, hamilton=False)

        cpd_range_dict = {}
        for cpd in ['A[c]', 'B[c]', 'C[c]']:
            cpd_range_dict[cpd] = \
                (tmfa.get_var_bound(prob, self.empty_xij[cpd],
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(prob, self.empty_xij[cpd],
                                    lp.ObjectiveSense.Maximize))

        self.assertAlmostEqual(cpd_range_dict['A[c]'][0],
                               (-5.298317366548036,
                                -5.298317366548036)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['A[c]'][1],
                               (-5.298317366548036,
                                -5.298317366548036)[1], places=6)
        self.assertAlmostEqual(cpd_range_dict['B[c]'][0],
                               (-7.600902459542082,
                                -5.298317366548036)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['B[c]'][1],
                               (-7.600902459542082,
                                -5.298317366548036)[1], places=6)
        self.assertAlmostEqual(cpd_range_dict['C[c]'][0],
                               (-11.512925026140119,
                                -3.912023005428146)[0], places=6)
        self.assertAlmostEqual(cpd_range_dict['C[c]'][1],
                               (-11.512925026140119,
                                -3.912023005428146)[1], places=6)

    def test_solvebiomass(self):
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.assertEqual(x, 1000)

    def test_rxnfluxes(self):
        flux_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for i in [i for i in self.mm_irreversible.reactions]:
            x = tmfa.get_var_bound(self.prob, self.v(i),
                                   lp.ObjectiveSense.Maximize)
            flux_dict[i] = x
        self.assertTrue(flux_dict['bio'] == 1000)
        self.assertTrue(flux_dict['rxn7'] == 0)
        self.assertTrue(flux_dict['rxn1_forward'] == 0)
        self.assertTrue(flux_dict['rxn4'] == 1000)

    def test_dgrrange(self):
        dgr_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn4', 'rxn6']:
            dgr_dict[reaction] = \
                (tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Maximize))

        self.assertAlmostEqual(dgr_dict['rxn1_forward'][0],
                               (9.999999974752427e-07,
                                69.28332553115047)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][1],
                               (9.999999974752427e-07,
                                69.28332553115047)[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][0],
                               (-39.64166326557521,
                                -9.999999903698154e-07)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][1],
                               (-39.64166326557521,
                                -9.999999903698154e-07)[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][0],
                               (265.3583367344248,
                                334.6416632655752)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][1],
                               (265.3583367344248,
                                334.6416632655752)[1], places=6)

    def test_cpd_range(self):
        cpd_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for cpd in ['A[c]', 'A[e]', 'D[c]', 'E[c]']:
            cpd_dict[cpd] = \
                (tmfa.get_var_bound(self.prob, self.xij(cpd),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.xij(cpd),
                                    lp.ObjectiveSense.Maximize))
        self.assertAlmostEqual(cpd_dict['A[c]'][1],
                               (-11.512925464970229,
                                -3.912023005428146)[1], places=6)
        self.assertAlmostEqual(cpd_dict['A[e]'][1],
                               (-11.512925464970229,
                                -3.912023005428146)[1], places=6)
        self.assertAlmostEqual(cpd_dict['D[c]'][1],
                               (-11.512925464970229,
                                -3.912023005428146)[1], places=6)

    def test_constrainedconc(self):
        cpd_dict = {}
        self.prob.add_linear_constraints(self.xij('A[c]') == -10)
        self.prob.add_linear_constraints(self.xij('B[c]') == -6)
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for cpd in ['A[c]', 'A[e]', 'D[c]', 'E[c]', 'B[c]']:
            cpd_dict[cpd] = \
                (tmfa.get_var_bound(self.prob, self.xij(cpd),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.xij(cpd),
                                    lp.ObjectiveSense.Maximize))
        dgr_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn4', 'rxn6']:
            dgr_dict[reaction] = \
                (tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertAlmostEqual(cpd_dict['D[c]'][1],
                               (-11.51292546497023,
                                -4.90292494313994)[1], places=6)
        self.assertAlmostEqual(cpd_dict['A[c]'][1],
                               (-10.0, -10.0)[1], places=6)
        self.assertAlmostEqual(cpd_dict['B[c]'][1],
                               (-6.0, -6.0)[1], places=6)
        self.assertAlmostEqual(cpd_dict['E[c]'][1],
                               (-11.512925245555174,
                                -3.912023005428146)[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][1],
                               (-30.125556943039474,
                                -9.999999974752427e-07)[1], places=6)

    def test_reversibledgri(self):
        dgr_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn1_reverse',
                         'rxn2_forward', 'rxn2_reverse']:
            dgr_dict[reaction] = \
                (tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.dgri(reaction),
                                    lp.ObjectiveSense.Maximize))

        self.assertAlmostEqual(dgr_dict['rxn1_forward'][1],
                               (dgr_dict['rxn1_reverse'][0],
                                dgr_dict['rxn1_reverse'][1])[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn2_forward'][1],
                               (dgr_dict['rxn2_reverse'][0],
                                dgr_dict['rxn2_reverse'][1])[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][1],
                               (dgr_dict['rxn1_reverse'][0],
                                dgr_dict['rxn1_reverse'][1])[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn2_forward'][1],
                               (dgr_dict['rxn2_reverse'][0],
                                dgr_dict['rxn2_reverse'][1])[1], places=6)

    def test_reversibleflux(self):
        flux_dict = {}
        x = tmfa.get_var_bound(self.prob, self.v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.prob.add_linear_constraints(self.v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn1_reverse',
                         'rxn2_forward', 'rxn2_reverse']:
            flux_dict[reaction] = \
                (tmfa.get_var_bound(self.prob, self.v(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.prob, self.v(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertTrue(flux_dict['rxn1_forward'] == (0, 0))
        self.assertTrue(flux_dict['rxn1_reverse'] == (1000, 1000))
        self.assertTrue(flux_dict['rxn2_forward'] == (1000, 1000))
        self.assertTrue(flux_dict['rxn2_reverse'] == (0, 0))

    def test_ham_flux(self):
        flux_dict = {}
        x = tmfa.get_var_bound(self.ham_prob, self.ham_v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.ham_prob.add_linear_constraints(self.ham_v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn2_forward', 'rxn2_reverse']:
            flux_dict[reaction] = \
                (tmfa.get_var_bound(self.ham_prob, self.ham_v(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.ham_prob, self.ham_v(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertTrue(flux_dict['rxn1_forward'] == (0, 0))
        self.assertTrue(flux_dict['rxn2_forward'] == (1000, 1000))
        self.assertTrue(flux_dict['rxn2_reverse'] == (0, 0))
        dgr_dict = {}
        for reaction in ['rxn1_forward', 'rxn4', 'rxn6']:
            dgr_dict[reaction] = \
                (tmfa.get_var_bound(self.ham_prob, self.ham_dgri(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.ham_prob, self.ham_dgri(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][0],
                               (9.999999974752427e-07,
                                69.28332553115047)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][1],
                               (9.999999974752427e-07,
                                69.28332553115047)[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][0],
                               (-39.64166326557521,
                                -9.999999903698154e-07)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][1],
                               (-39.64166326557521,
                                -9.999999903698154e-07)[1], places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][0],
                               (265.3583377344248, 299.999999)[0], places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][1],
                               (265.3583377344248, 299.999999)[1], places=6)

    def test_err_flux(self):
        flux_dict = {}
        x = tmfa.get_var_bound(self.err_prob, self.err_v('bio'),
                               lp.ObjectiveSense.Maximize)
        self.err_prob.add_linear_constraints(self.err_v('bio') == x)
        for reaction in ['rxn1_forward', 'rxn2_forward', 'rxn2_reverse']:
            flux_dict[reaction] = \
                (tmfa.get_var_bound(self.err_prob, self.err_v(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.err_prob, self.err_v(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertTrue(flux_dict['rxn1_forward'] == (0, 0))
        self.assertTrue(flux_dict['rxn2_forward'] == (1000, 1000))
        self.assertTrue(flux_dict['rxn2_reverse'] == (0, 0))
        dgr_dict = {}
        for reaction in ['rxn1_forward', 'rxn4', 'rxn6']:
            dgr_dict[reaction] = \
                (tmfa.get_var_bound(self.err_prob, self.err_dgri(reaction),
                                    lp.ObjectiveSense.Minimize),
                 tmfa.get_var_bound(self.err_prob, self.err_dgri(reaction),
                                    lp.ObjectiveSense.Maximize))
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][0],
                               (-7.999998999999999, 77.28332553115047)[0],
                               places=6)
        self.assertAlmostEqual(dgr_dict['rxn1_forward'][1],
                               (9.999999974752427e-07, 77.28332553115047)[1],
                               places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][0],
                               (-43.64166326557521, -9.999999903698154e-07)[0],
                               places=6)
        self.assertAlmostEqual(dgr_dict['rxn4'][1],
                               (-43.64166326557521, -9.999999903698154e-07)[1],
                               places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][0],
                               (261.3583367344248, 338.6416632655752)[0],
                               places=6)
        self.assertAlmostEqual(dgr_dict['rxn6'][1],
                               (261.3583367344248, 338.6416632655752)[1],
                               places=6)
